import asyncio
from typing import TYPE_CHECKING, Dict, Optional

import websockets
from websockets.server import WebSocketServerProtocol as WsClient

from arcor2.data import rpc

from arcor2.server import globals as glob


if TYPE_CHECKING:
    ReqQueue = asyncio.Queue[rpc.common.Request]
    RespQueue = asyncio.Queue[rpc.common.Response]
else:
    ReqQueue = asyncio.Queue
    RespQueue = asyncio.Queue

MANAGER_RPC_REQUEST_QUEUE: ReqQueue = ReqQueue()
MANAGER_RPC_RESPONSES: Dict[int, RespQueue] = {}


async def manager_request(req: rpc.common.Request, ui: Optional[WsClient] = None) -> rpc.common.Response:

    assert req.id not in MANAGER_RPC_RESPONSES

    MANAGER_RPC_RESPONSES[req.id] = RespQueue(maxsize=1)
    await MANAGER_RPC_REQUEST_QUEUE.put(req)

    resp = await MANAGER_RPC_RESPONSES[req.id].get()
    del MANAGER_RPC_RESPONSES[req.id]
    return resp


async def project_manager_client(handle_manager_incoming_messages) -> None:

    while True:

        await glob.logger.info("Attempting connection to manager...")

        try:

            async with websockets.connect(glob.MANAGER_URL) as manager_client:

                await glob.logger.info("Connected to manager.")

                future = asyncio.ensure_future(handle_manager_incoming_messages(manager_client))

                while True:

                    if future.done():
                        break

                    try:
                        msg = await asyncio.wait_for(MANAGER_RPC_REQUEST_QUEUE.get(), 1.0)
                    except asyncio.TimeoutError:
                        continue

                    try:
                        await manager_client.send(msg.to_json())
                    except websockets.exceptions.ConnectionClosed:
                        await MANAGER_RPC_REQUEST_QUEUE.put(msg)
                        break
        except ConnectionRefusedError as e:
            await glob.logger.error(e)
            await asyncio.sleep(delay=1.0)



# -*- coding: utf-8 -*-

import re
from dataclasses import dataclass, field
from typing import List, Optional

from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.data import common, execution


def wo_suffix(name: str) -> str:
    return re.sub('Event$', '', name)


class EventType(common.StrEnum):

    ADD: str = "add"
    UPDATE: str = "update"
    REMOVE: str = "remove"
    UPDATE_BASE: str = "update_base"


"""
------------------------------------------------------------------------------------------------------------------------
Common stuff
------------------------------------------------------------------------------------------------------------------------
"""


@dataclass
class Event(JsonSchemaMixin):

    event: str = field(default="", init=False)
    change_type: Optional[EventType] = None
    parent_id: Optional[str] = None


"""
------------------------------------------------------------------------------------------------------------------------
Project / scene
------------------------------------------------------------------------------------------------------------------------
"""


@dataclass
class OpenSceneData(JsonSchemaMixin):

    scene: common.Scene


@dataclass
class OpenScene(Event):

    data: Optional[OpenSceneData] = None
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class SceneChanged(Event):

    data: Optional[common.Scene] = None
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class SceneSaved(Event):

    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class SceneClosed(Event):

    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ProjectChanged(Event):

    data: Optional[common.Project] = None
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class OpenProjectData(JsonSchemaMixin):

    scene: common.Scene
    project: common.Project


@dataclass
class OpenProject(Event):

    data: Optional[OpenProjectData] = None
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ProjectSaved(Event):

    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ProjectClosed(Event):

    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ActionPointChanged(Event):

    data: Optional[common.ProjectActionPoint] = None
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ActionChanged(Event):

    data: Optional[common.Action] = None
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class OrientationChanged(Event):

    data: Optional[common.NamedOrientation] = None
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class JointsChanged(Event):

    data: Optional[common.ProjectRobotJoints] = None
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class SceneObjectChanged(Event):

    data: Optional[common.SceneObject] = None
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class SceneServiceChanged(Event):

    data: Optional[common.SceneService] = None
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


"""
------------------------------------------------------------------------------------------------------------------------
Project execution
------------------------------------------------------------------------------------------------------------------------
"""


@dataclass
class ProjectExceptionEventData(JsonSchemaMixin):

    message: str = field(default_factory=str)
    type: str = field(default_factory=str)
    handled: bool = False


@dataclass
class ProjectExceptionEvent(Event):

    data: ProjectExceptionEventData = field(default_factory=ProjectExceptionEventData)
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class CurrentActionEvent(Event):

    data: common.CurrentAction = field(default_factory=common.CurrentAction)
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class PackageStateEvent(Event):

    data: common.PackageState = field(default_factory=common.PackageState)
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class PackageInfoEvent(Event):

    data: Optional[execution.PackageInfo] = None
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


# ----------------------------------------------------------------------------------------------------------------------

@dataclass
class ActionStateEvent(Event):

    data: common.ActionState = field(default_factory=common.ActionState)
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ActionExecutionData(JsonSchemaMixin):

    action_id: str = ""


@dataclass
class ActionExecutionEvent(Event):
    """
    Used to notify beginning of the action execution triggered while editing a project.
    """

    data: ActionExecutionData = field(default_factory=ActionExecutionData)
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ActionResult(JsonSchemaMixin):

    action_id: str = ""
    result: Optional[str] = field(default=None, metadata=dict(description="JSON-encoded result of the action."))
    error: Optional[str] = None


@dataclass
class ActionResultEvent(Event):
    """
    Used to notify result of the action execution triggered while editing a project.
    """

    data: ActionResult = field(default_factory=ActionResult)
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ActionCancelledEvent(Event):
    """
    Used to notify that the action execution was cancelled.
    """

    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


"""
------------------------------------------------------------------------------------------------------------------------
Objects
------------------------------------------------------------------------------------------------------------------------
"""


@dataclass
class ObjectTypesChangedEvent(Event):

    data: List[str] = field(default_factory=list)  # changed object types
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


"""
------------------------------------------------------------------------------------------------------------------------
Robots
------------------------------------------------------------------------------------------------------------------------
"""


@dataclass
class RobotJointsData(JsonSchemaMixin):

    robot_id: str
    joints: List[common.Joint]


@dataclass
class RobotJointsEvent(Event):

    data: Optional[RobotJointsData] = None
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class EefPose(JsonSchemaMixin):

    end_effector_id: str
    pose: common.Pose


@dataclass
class RobotEefData(JsonSchemaMixin):

    robot_id: str
    end_effectors: List[EefPose] = field(default_factory=list)


@dataclass
class RobotEefEvent(Event):

    data: Optional[RobotEefData] = None
    event: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821



from arcor2.data.rpc import common
from arcor2.data.rpc import execution
from arcor2.data.rpc import objects
from arcor2.data.rpc import robot
from arcor2.data.rpc import scene
from arcor2.data.rpc import project
from arcor2.data.rpc import services
from arcor2.data.rpc import storage

__all__ = ["common", "execution", "objects", "robot", "scene", "project", "services", "storage"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from dataclasses import dataclass, field
from datetime import datetime, timezone
from enum import Enum, unique
from json import JSONEncoder
from typing import List, Any, Iterator, Optional, Tuple, Set, Union
import uuid

import quaternion  # type: ignore
from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.exceptions import Arcor2Exception


def uid() -> str:
    return uuid.uuid4().hex


@unique
class StrEnum(str, Enum):

    @classmethod
    def set(cls) -> Set[str]:
        return set(map(lambda c: c.value, cls))  # type: ignore


@unique
class IntEnum(int, Enum):

    @classmethod
    def set(cls) -> Set[int]:
        return set(map(lambda c: c.value, cls))  # type: ignore


class ActionIOEnum(StrEnum):

    FIRST: str = "start"
    LAST: str = "end"


class DataClassEncoder(JSONEncoder):

    def default(self, o: Any) -> Any:

        if isinstance(o, JsonSchemaMixin):
            return o.to_dict()

        return super().default(o)


@dataclass
class IterableIndexable(JsonSchemaMixin):

    def __getitem__(self, item: int) -> float:

        attr = getattr(self, tuple(self.__dict__.keys())[item])
        assert isinstance(attr, float)
        return attr

    def __iter__(self) -> Iterator[float]:

        yield from self.__dict__.values()


@dataclass
class Position(IterableIndexable):

    x: float = 0.0
    y: float = 0.0
    z: float = 0.0

    def rotated(self, rot: "Orientation", inverse: bool = False) -> "Position":

        q = rot.as_quaternion()

        if inverse:
            q = q.inverse()

        rotated_vector = quaternion.rotate_vectors([q], [list(self)])[0][0]
        return Position(rotated_vector[0], rotated_vector[1], rotated_vector[2])


@dataclass
class Orientation(IterableIndexable):

    x: float = 0.0
    y: float = 0.0
    z: float = 0.0
    w: float = 1.0

    def as_quaternion(self) -> quaternion.quaternion:
        return quaternion.quaternion(self.w, self.x, self.y, self.z).normalized()

    def set_from_quaternion(self, q: quaternion.quaternion) -> None:

        nq = q.normalized()

        self.x = nq.x
        self.y = nq.y
        self.z = nq.z
        self.w = nq.w

    def __eq__(self, other) -> bool:

        if not isinstance(other, Orientation):
            return False

        return quaternion.isclose(self.as_quaternion(), other.as_quaternion(), rtol=1.e-8)[0]


@dataclass
class NamedOrientation(JsonSchemaMixin):

    id: str
    name: str
    orientation: Orientation


@dataclass
class Pose(JsonSchemaMixin):

    position: Position = field(default_factory=Position)
    orientation: Orientation = field(default_factory=Orientation)


@dataclass
class ActionMetadata(JsonSchemaMixin):

    free: bool = False
    blocking: bool = False
    composite: bool = False
    blackbox: bool = False
    cancellable: bool = field(init=False, default=False)


@dataclass
class Joint(JsonSchemaMixin):

    name: str
    value: float


@dataclass
class ProjectRobotJoints(JsonSchemaMixin):

    id: str
    name: str
    robot_id: str
    joints: List[Joint]
    is_valid: bool = False


@dataclass
class ActionPoint(JsonSchemaMixin):

    id: str
    name: str
    position: Position
    parent: Optional[str] = None
    orientations: List[NamedOrientation] = field(default_factory=list)
    robot_joints: List[ProjectRobotJoints] = field(default_factory=list)

    def pose(self, orientation_id: str = "default") -> Pose:

        for ori in self.orientations:
            if ori.id == orientation_id:
                return Pose(self.position, ori.orientation)
        raise Arcor2Exception(f"Action point {self.id} does not contain orientation {orientation_id}.")

    def orientation(self, orientation_id: str) -> NamedOrientation:

        for ori in self.orientations:
            if ori.id == orientation_id:
                return ori
        raise Arcor2Exception(f"Action point {self.id} does not contain orientation {orientation_id}.")

    def joints(self, joints_id: str) -> ProjectRobotJoints:

        for joints in self.robot_joints:
            if joints.id == joints_id:
                return joints
        raise Arcor2Exception(f"Action point {self.id} does not contain robot joints {joints_id}.")

    def orientation_names(self) -> Set[str]:
        return {ori.name for ori in self.orientations}

    def joints_names(self) -> Set[str]:
        return {joints.name for joints in self.robot_joints}

    def joints_for_robot(self, robot_id: str, joints_id: str) -> ProjectRobotJoints:

        joints = self.joints(joints_id)

        if joints.robot_id != robot_id:
            raise Arcor2Exception("Joints for a different robot.")

        return joints

    def invalidate_joints(self):

        for joints in self.robot_joints:
            joints.is_valid = False


@dataclass
class SceneObject(JsonSchemaMixin):

    id: str
    name: str
    type: str
    pose: Pose


@dataclass
class SceneService(JsonSchemaMixin):

    type: str
    configuration_id: str


@dataclass
class Scene(JsonSchemaMixin):

    id: str
    name: str
    objects: List[SceneObject] = field(default_factory=list)
    services: List[SceneService] = field(default_factory=list)
    desc: str = field(default_factory=str)
    modified: Optional[datetime] = None
    int_modified: Optional[datetime] = None

    def bare(self) -> "Scene":
        return Scene(self.id, self.name, desc=self.desc)

    def update_modified(self):
        self.int_modified = datetime.now(tz=timezone.utc)

    def has_changes(self) -> bool:

        if self.int_modified is None:
            return False

        if self.modified is None:
            return True

        return self.int_modified > self.modified

    def object_names(self) -> Iterator[str]:

        for obj in self.objects:
            yield obj.name

    def object(self, object_id: str) -> SceneObject:

        for obj in self.objects:
            if obj.id == object_id:
                return obj
        raise Arcor2Exception(f"Object ID {object_id} not found.")

    def service(self, service_type: str) -> SceneService:

        for srv in self.services:
            if srv.type == service_type:
                return srv
        raise Arcor2Exception(f"Service of type {service_type} not found.")

    def object_or_service(self, object_or_service_id: str) -> Union[SceneObject, SceneService]:

        try:
            return self.object(object_or_service_id)
        except Arcor2Exception:
            pass

        try:
            return self.service(object_or_service_id)
        except Arcor2Exception:
            pass

        raise Arcor2Exception(f"Scene does not contain object/service with id '{object_or_service_id}'.")


@dataclass
class IdValue(JsonSchemaMixin):

    id: str
    value: str


class ActionParameterException(Arcor2Exception):
    pass


@dataclass
class ActionParameter(IdValue):

    type: str


@dataclass
class ActionIO(JsonSchemaMixin):

    default: str


@dataclass
class Action(JsonSchemaMixin):

    id: str
    name: str
    type: str
    parameters: List[ActionParameter] = field(default_factory=list)
    inputs: List[ActionIO] = field(default_factory=list)
    outputs: List[ActionIO] = field(default_factory=list)

    def parse_type(self) -> Tuple[str, str]:

        try:
            obj_id_str, action = self.type.split("/")
        except ValueError:
            raise Arcor2Exception(f"Action: {self.id} has invalid type: {self.type}.")
        return obj_id_str, action

    def parameter(self, parameter_id) -> ActionParameter:

        for param in self.parameters:
            if parameter_id == param.id:
                return param

        raise Arcor2Exception("Param not found")

    def bare(self) -> "Action":
        return Action(self.id, self.name, self.type)


@dataclass
class ProjectActionPoint(ActionPoint):

    actions: List[Action] = field(default_factory=list)

    def bare(self) -> "ProjectActionPoint":
        return ProjectActionPoint(self.id, self.name, self.position, self.parent)


@dataclass
class Project(JsonSchemaMixin):

    id: str
    name: str
    scene_id: str
    action_points: List[ProjectActionPoint] = field(default_factory=list)
    desc: str = field(default_factory=str)
    has_logic: bool = True
    modified: Optional[datetime] = None
    int_modified: Optional[datetime] = None

    def bare(self) -> "Project":
        return Project(self.id, self.name, self.scene_id, desc=self.desc, has_logic=self.has_logic)

    def update_modified(self):
        self.int_modified = datetime.now(tz=timezone.utc)

    def has_changes(self) -> bool:

        if self.int_modified is None:
            return False

        if self.modified is None:
            return True

        return self.int_modified > self.modified

    @property
    def action_points_with_parent(self) -> List[ProjectActionPoint]:
        """
        Get action points which are relative to something (parent is set).
        :return:
        """

        return [ap for ap in self.action_points if ap.parent]

    @property
    def action_points_names(self) -> Set[str]:
        return {ap.name for ap in self.action_points}

    def ap_and_joints(self, joints_id: str) -> Tuple[ProjectActionPoint, ProjectRobotJoints]:

        for ap in self.action_points:
            for joints in ap.robot_joints:
                if joints.id == joints_id:
                    return ap, joints
        raise Arcor2Exception("Unknown joints.")

    def joints(self, joints_id: str) -> ProjectRobotJoints:
        return self.ap_and_joints(joints_id)[1]

    def ap_and_orientation(self, orientation_id: str) -> Tuple[ProjectActionPoint, NamedOrientation]:

        for ap in self.action_points:
            for ori in ap.orientations:
                if ori.id == orientation_id:
                    return ap, ori
        raise Arcor2Exception("Unknown orientation.")

    def orientation(self, orientation_id: str) -> NamedOrientation:
        return self.ap_and_orientation(orientation_id)[1]

    def action(self, action_id: str) -> Action:

        for ap in self.action_points:
            for ac in ap.actions:
                if ac.id == action_id:
                    return ac
        else:
            raise Arcor2Exception("Action not found")

    def action_point_and_action(self, action_id: str) -> Tuple[ProjectActionPoint, Action]:

        for ap in self.action_points:
            for ac in ap.actions:
                if ac.id == action_id:
                    return ap, ac
        else:
            raise Arcor2Exception("Action not found")

    def actions(self) -> List[Action]:
        return [act for ap in self.action_points for act in ap.actions]

    def action_ids(self) -> Set[str]:
        return {action.id for action in self.actions()}

    def action_user_names(self) -> Set[str]:
        return {action.name for action in self.actions()}

    def action_point(self, action_point_id: str) -> ProjectActionPoint:

        for ap in self.action_points:
            if ap.id == action_point_id:
                return ap
        else:
            raise Arcor2Exception("Action point not found")


@dataclass
class ProjectSources(JsonSchemaMixin):

    id: str  # project_id
    script: str


@dataclass
class IdDesc(JsonSchemaMixin):
    id: str
    name: str
    desc: Optional[str] = None


@dataclass
class IdDescList(JsonSchemaMixin):

    items: List[IdDesc] = field(default_factory=list)


class PackageStateEnum(Enum):

    RUNNING: str = "running"
    STOPPED: str = "stopped"
    PAUSED: str = "paused"


class ActionStateEnum(Enum):

    BEFORE: str = "before"
    AFTER: str = "after"


@dataclass
class ActionState(JsonSchemaMixin):

    object_id: str = ""
    method: str = ""
    where: ActionStateEnum = ActionStateEnum.BEFORE


@dataclass
class PackageState(JsonSchemaMixin):

    state: PackageStateEnum = PackageStateEnum.STOPPED


@dataclass
class CurrentAction(JsonSchemaMixin):

    action_id: str = ""
    args: List[ActionParameter] = field(default_factory=list)


@dataclass
class BroadcastInfo(JsonSchemaMixin):

    host: str
    port: int



from arcor2.server.rpc import execution
from arcor2.server.rpc import objects
from arcor2.server.rpc import project
from arcor2.server.rpc import robot
from arcor2.server.rpc import scene
from arcor2.server.rpc import services

__all__ = ["execution", "objects", "project", "robot", "scene", "services"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2.data import rpc
from arcor2.server import globals as glob


async def get_services_cb(req: rpc.services.GetServicesRequest, ui: WsClient) -> rpc.services.GetServicesResponse:
    return rpc.services.GetServicesResponse(data=list(glob.SERVICE_TYPES.values()))



import inspect
from typing import Type

from arcor2.source.services import check_service_type
from arcor2.source import SourceException
from arcor2.helpers import import_cls, ImportClsException
from arcor2.exceptions import Arcor2Exception
from arcor2 import persistent_storage as storage
from arcor2.data.services import ServiceType
from arcor2.services import Service


class UploadException(Arcor2Exception):
    pass


def upload_cls(cls: Type[Service]) -> None:

    path = inspect.getfile(cls)

    with open(path, "r") as source_file:

        source = source_file.read()

        try:
            check_service_type(source, cls.__name__)
        except SourceException as e:
            print(e)
            raise UploadException(f"There is something wrong with source code of '{cls.__name__}'.")

        srv_type = ServiceType(id=cls.__name__, source=source, desc=cls.description())

        print(f"Storing '{srv_type.id}'...")
        storage.update_service_type(srv_type)


def upload(module_cls: str) -> None:

    try:
        _, cls = import_cls(module_cls)
    except ImportClsException as e:
        raise UploadException(e)

    if not issubclass(cls, Service):
        print("Not subclass of Service!")
        return

    upload_cls(cls)



import abc
from typing import FrozenSet, List, Optional

from arcor2.object_types.generic import Generic
from arcor2.data.common import Pose, Joint


class Robot(Generic, metaclass=abc.ABCMeta):
    """
    Abstract class representing robot and its basic capabilities (motion)
    """

    urdf_package_path: Optional[str] = None

    @abc.abstractmethod
    def get_end_effectors_ids(self) -> FrozenSet[str]:
        pass

    @abc.abstractmethod
    def get_end_effector_pose(self, end_effector: str) -> Pose:
        pass

    @abc.abstractmethod
    def robot_joints(self) -> List[Joint]:
        pass

    @abc.abstractmethod
    def grippers(self) -> FrozenSet[str]:
        return frozenset()

    @abc.abstractmethod
    def suctions(self) -> FrozenSet[str]:
        return frozenset()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import asyncio
from typing import Dict, Any, Set
import copy
from contextlib import asynccontextmanager
import inspect

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import object_types_utils as otu, helpers as hlp
from arcor2.data import rpc, events, common, object_type
from arcor2 import aio_persistent_storage as storage
from arcor2.exceptions import Arcor2Exception
from arcor2.parameter_plugins import PARAM_PLUGINS
from arcor2.parameter_plugins.base import ParameterPluginException
from arcor2.source.logic import program_src, SourceException

from arcor2.server.decorators import scene_needed, project_needed, no_project
from arcor2.server import objects_services_actions as osa, notifications as notif, globals as glob
from arcor2.server.robot import get_end_effector_pose, get_robot_joints
from arcor2.server.project import project_problems, open_project, project_names
from arcor2.server.scene import open_scene, clear_scene, get_instance


def find_object_action(action: common.Action) -> object_type.ObjectAction:

    assert glob.SCENE

    obj_id, action_type = action.parse_type()
    obj = glob.SCENE.object_or_service(obj_id)

    if obj.type not in glob.ACTIONS:
        raise Arcor2Exception("Unknown object/service type.")

    for act in glob.ACTIONS[obj.type]:
        if act.name == action_type:
            if act.disabled:
                raise Arcor2Exception("Action is disabled.")
            return act
    raise Arcor2Exception("Unknown type of action.")


@asynccontextmanager
async def managed_project(project_id: str, make_copy: bool = False):

    save_back = False

    if glob.PROJECT and glob.PROJECT.id == project_id:
        if make_copy:
            project = copy.deepcopy(glob.PROJECT)
            save_back = True
        else:
            project = glob.PROJECT
    else:
        save_back = True
        project = await storage.get_project(project_id)

    if make_copy:
        project.id = common.uid()

    try:
        yield project
    finally:
        if save_back:
            asyncio.ensure_future(storage.update_project(project))


def unique_name(name: str, existing_names: Set[str]) -> None:

    if not name:
        raise Arcor2Exception("Name has to be set.")

    if name in existing_names:
        raise Arcor2Exception("Name already exists.")


@scene_needed
@project_needed
async def cancel_action_cb(req: rpc.project.CancelActionRequest, ui: WsClient) -> None:

    assert glob.PROJECT

    if not glob.RUNNING_ACTION:
        raise Arcor2Exception("No action is running.")

    action = glob.PROJECT.action(glob.RUNNING_ACTION)
    obj_id, action_name = action.parse_type()
    obj = get_instance(obj_id)

    for act in glob.ACTIONS[obj.__class__.__name__]:
        if act.name != action_name:
            continue

        if not act.meta.cancellable:
            raise Arcor2Exception("Action is not cancellable.")

    try:
        action_method = getattr(obj, action_name)
        cancel_method = getattr(obj, obj.CANCEL_MAPPING[action_method.__name__])
    except AttributeError as e:
        raise Arcor2Exception("Internal error.") from e

    cancel_params: Dict[str, Any] = {}

    cancel_sig = inspect.signature(cancel_method)

    assert glob.RUNNING_ACTION_PARAMS is not None

    for param_name, param in cancel_sig.parameters.items():
        try:
            cancel_params[param_name] = glob.RUNNING_ACTION_PARAMS[param_name]
        except KeyError as e:
            raise Arcor2Exception("Cancel method parameters should be subset of action parameters.") from e

    await hlp.run_in_executor(cancel_method, *cancel_params.values())

    asyncio.ensure_future(notif.broadcast_event(events.ActionCancelledEvent()))
    glob.RUNNING_ACTION = None
    glob.RUNNING_ACTION_PARAMS = None


@scene_needed
@project_needed
async def execute_action_cb(req: rpc.project.ExecuteActionRequest, ui: WsClient) -> None:

    assert glob.SCENE and glob.PROJECT

    if glob.RUNNING_ACTION:
        raise Arcor2Exception(f"Action {glob.RUNNING_ACTION} is being executed. "
                              f"Only one action can be executed at a time.")

    action = glob.PROJECT.action(req.args.action_id)

    params: Dict[str, Any] = {}

    for param in action.parameters:
        try:
            params[param.id] = PARAM_PLUGINS[param.type].execution_value(glob.TYPE_DEF_DICT, glob.SCENE,
                                                                         glob.PROJECT, action.id, param.id)
        except ParameterPluginException as e:
            await glob.logger.error(e)
            raise Arcor2Exception(f"Failed to get value for parameter {param.id}.")

    obj_id, action_name = action.parse_type()

    obj = get_instance(obj_id)

    if not hasattr(obj, action_name):
        raise Arcor2Exception("Internal error: object does not have the requested method.")

    glob.RUNNING_ACTION = action.id
    glob.RUNNING_ACTION_PARAMS = params

    await glob.logger.debug(f"Running action {action.name} ({type(obj)}/{action_name}), params: {params}.")

    # schedule execution and return success
    asyncio.ensure_future(osa.execute_action(getattr(obj, action_name), params))
    return None


async def project_info(project_id: str, scenes_lock: asyncio.Lock, scenes: Dict[str, common.Scene]) -> \
        rpc.project.ListProjectsResponseData:

    project = await storage.get_project(project_id)

    pd = rpc.project.ListProjectsResponseData(id=project.id, desc=project.desc, name=project.name,
                                              scene_id=project.scene_id)

    try:
        async with scenes_lock:
            if project.scene_id not in scenes:
                scenes[project.scene_id] = await storage.get_scene(project.scene_id)
    except storage.PersistentStorageException:
        pd.problems.append("Scene does not exist.")
        return pd

    pd.problems = project_problems(scenes[project.scene_id], project)
    pd.valid = not pd.problems

    if not pd.valid:
        return pd

    try:
        program_src(project, scenes[project.scene_id], otu.built_in_types_names())
        pd.executable = True
    except SourceException as e:
        pd.problems.append(e.message)

    return pd


async def list_projects_cb(req: rpc.project.ListProjectsRequest, ui: WsClient) -> rpc.project.ListProjectsResponse:

    projects = await storage.get_projects()

    scenes_lock = asyncio.Lock()
    scenes: Dict[str, common.Scene] = {}

    resp = rpc.project.ListProjectsResponse()
    tasks = [project_info(project_iddesc.id, scenes_lock, scenes) for project_iddesc in projects.items]
    resp.data = await asyncio.gather(*tasks)
    return resp


@scene_needed
@project_needed
async def add_action_point_joints_cb(req: rpc.project.AddActionPointJointsRequest, ui: WsClient) -> None:

    assert glob.SCENE and glob.PROJECT

    ap = glob.PROJECT.action_point(req.args.action_point_id)

    unique_name(req.args.name, ap.joints_names())

    if req.dry_run:
        return None

    new_joints = await get_robot_joints(req.args.robot_id)

    prj = common.ProjectRobotJoints(common.uid(), req.args.name, req.args.robot_id, new_joints, True)
    ap.robot_joints.append(prj)
    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.JointsChanged(events.EventType.ADD, ap.id, data=prj)))
    return None


@scene_needed
@project_needed
async def update_action_point_joints_cb(req: rpc.project.UpdateActionPointJointsRequest, ui: WsClient) -> None:

    assert glob.SCENE and glob.PROJECT

    robot_joints = glob.PROJECT.joints(req.args.joints_id)
    new_joints = await get_robot_joints(req.args.robot_id)
    robot_joints.joints = new_joints
    robot_joints.robot_id = req.args.robot_id
    robot_joints.is_valid = True

    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.JointsChanged(events.EventType.UPDATE, data=robot_joints)))
    return None


@scene_needed
@project_needed
async def remove_action_point_joints_cb(req: rpc.project.RemoveActionPointJointsRequest, ui: WsClient) -> None:
    """
    Removes joints from action point.
    :param req:
    :return:
    """

    assert glob.SCENE and glob.PROJECT

    ap, joints = glob.PROJECT.ap_and_joints(req.args.joints_id)

    for act in glob.PROJECT.actions():
        for param in act.parameters:
            if PARAM_PLUGINS[param.type].uses_robot_joints(glob.PROJECT, act.id, param.id, req.args.joints_id):
                raise Arcor2Exception(f"Joints used in action {act.name} (parameter {param.id}).")

    joints_to_be_removed = ap.joints(req.args.joints_id)
    ap.robot_joints = [joints for joints in ap.robot_joints if joints.id != req.args.joints_id]

    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.JointsChanged(events.EventType.REMOVE,
                                                                     data=joints_to_be_removed)))
    return None


@scene_needed
@project_needed
async def rename_action_point_cb(req: rpc.project.RenameActionPointRequest, ui: WsClient) -> None:

    assert glob.SCENE and glob.PROJECT

    ap = glob.PROJECT.action_point(req.args.action_point_id)

    if req.args.new_name == ap.name:
        return None

    if not hlp.is_valid_identifier(req.args.new_name):
        raise Arcor2Exception("Name has to be valid Python identifier.")

    unique_name(req.args.new_name, glob.PROJECT.action_points_names)

    if req.dry_run:
        return None

    ap.name = req.args.new_name

    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.ActionPointChanged(events.EventType.UPDATE_BASE,
                                                                          data=ap.bare())))
    return None


@scene_needed
@project_needed
async def update_action_point_parent_cb(req: rpc.project.UpdateActionPointParentRequest, ui: WsClient) -> None:

    assert glob.SCENE and glob.PROJECT

    ap = glob.PROJECT.action_point(req.args.action_point_id)

    if req.args.new_parent_id == ap.parent:
        return None

    if not ap.parent and req.args.new_parent_id:
        # AP position and all orientations will become relative to the parent
        new_parent = glob.SCENE.object(req.args.new_parent_id)
        ap.position = hlp.make_pose_rel(new_parent.pose, common.Pose(ap.position, common.Orientation())).position
        for ori in ap.orientations:
            ori.orientation = hlp.make_orientation_rel(new_parent.pose.orientation, ori.orientation)

    elif ap.parent and not req.args.new_parent_id:
        # AP position and all orientations will become absolute
        old_parent = glob.SCENE.object(ap.parent)
        ap.position = hlp.make_pose_abs(old_parent.pose, common.Pose(ap.position, common.Orientation())).position
        for ori in ap.orientations:
            ori.orientation = hlp.make_orientation_abs(old_parent.pose.orientation, ori.orientation)
    else:

        assert ap.parent is not None

        # AP position and all orientations will become relative to another parent
        old_parent = glob.SCENE.object(ap.parent)
        new_parent = glob.SCENE.object(req.args.new_parent_id)

        abs_ap_pose = hlp.make_pose_abs(old_parent.pose, common.Pose(ap.position, common.Orientation()))
        ap.position = hlp.make_pose_rel(new_parent.pose, abs_ap_pose).position

        for ori in ap.orientations:
            ori.orientation = hlp.make_orientation_abs(old_parent.pose.orientation, ori.orientation)
            ori.orientation = hlp.make_orientation_rel(new_parent.pose.orientation, ori.orientation)

    ap.parent = req.args.new_parent_id
    glob.PROJECT.update_modified()

    """
    Can't send orientation changes and then ActionPointChanged/UPDATE_BASE (or vice versa)
    because UI would display orientations wrongly (for a short moment).
    """
    asyncio.ensure_future(notif.broadcast_event(events.ActionPointChanged(events.EventType.UPDATE,
                                                                          data=ap)))
    return None


@scene_needed
@project_needed
async def update_action_point_position_cb(req: rpc.project.UpdateActionPointPositionRequest, ui: WsClient) -> None:

    assert glob.SCENE and glob.PROJECT

    ap = glob.PROJECT.action_point(req.args.action_point_id)

    ap.position = req.args.new_position
    ap.invalidate_joints()
    for joints in ap.robot_joints:
        asyncio.ensure_future(
            notif.broadcast_event(events.JointsChanged(events.EventType.UPDATE, ap.id, data=joints)))

    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.ActionPointChanged(events.EventType.UPDATE_BASE,
                                                                          data=ap.bare())))
    return None


@scene_needed
@project_needed
async def update_action_point_using_robot_cb(req: rpc.project.UpdateActionPointUsingRobotRequest, ui: WsClient) -> None:

    assert glob.SCENE and glob.PROJECT

    ap = glob.PROJECT.action_point(req.args.action_point_id)
    new_pose = await get_end_effector_pose(req.args.robot.robot_id, req.args.robot.end_effector)

    if ap.parent:

        obj = glob.SCENE_OBJECT_INSTANCES[ap.parent]
        new_pose = hlp.make_pose_rel(obj.pose, new_pose)

    ap.invalidate_joints()
    for joints in ap.robot_joints:
        asyncio.ensure_future(
            notif.broadcast_event(events.JointsChanged(events.EventType.UPDATE, ap.id, data=joints)))

    ap.position = new_pose.position

    glob.PROJECT.update_modified()

    asyncio.ensure_future(notif.broadcast_event(events.ActionPointChanged(events.EventType.UPDATE_BASE,
                                                                          data=ap.bare())))
    return None


@scene_needed
@project_needed
async def add_action_point_orientation_cb(req: rpc.project.AddActionPointOrientationRequest, ui: WsClient) -> None:
    """
    Adds orientation and joints to the action point.
    :param req:
    :return:
    """

    assert glob.SCENE and glob.PROJECT

    ap = glob.PROJECT.action_point(req.args.action_point_id)

    unique_name(req.args.name, ap.orientation_names())

    if req.dry_run:
        return None

    orientation = common.NamedOrientation(common.uid(), req.args.name, req.args.orientation)
    ap.orientations.append(orientation)

    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.OrientationChanged(events.EventType.ADD, ap.id,
                                                                          data=orientation)))
    return None


@scene_needed
@project_needed
async def update_action_point_orientation_cb(req: rpc.project.UpdateActionPointOrientationRequest, ui: WsClient) -> \
        None:
    """
    Updates orientation of the action point.
    :param req:
    :return:
    """

    assert glob.SCENE and glob.PROJECT

    orientation = glob.PROJECT.orientation(req.args.orientation_id)
    orientation.orientation = req.args.orientation

    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.OrientationChanged(events.EventType.UPDATE, data=orientation)))
    return None


@scene_needed
@project_needed
async def add_action_point_orientation_using_robot_cb(req: rpc.project.AddActionPointOrientationUsingRobotRequest,
                                                      ui: WsClient) -> None:
    """
    Adds orientation and joints to the action point.
    :param req:
    :return:
    """

    assert glob.SCENE and glob.PROJECT

    ap = glob.PROJECT.action_point(req.args.action_point_id)
    unique_name(req.args.name, ap.orientation_names())

    if req.dry_run:
        return None

    new_pose = await get_end_effector_pose(req.args.robot.robot_id, req.args.robot.end_effector)

    if ap.parent:
        obj = glob.SCENE_OBJECT_INSTANCES[ap.parent]
        new_pose = hlp.make_pose_rel(obj.pose, new_pose)

    orientation = common.NamedOrientation(common.uid(), req.args.name, new_pose.orientation)
    ap.orientations.append(orientation)

    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.OrientationChanged(events.EventType.ADD, ap.id,
                                                                          data=orientation)))
    return None


@scene_needed
@project_needed
async def update_action_point_orientation_using_robot_cb(
        req: rpc.project.UpdateActionPointOrientationUsingRobotRequest, ui: WsClient) -> None:
    """
    Updates orientation and joint of the action point.
    :param req:
    :return:
    """

    assert glob.SCENE and glob.PROJECT

    ap = glob.PROJECT.action_point(req.args.action_point_id)
    new_pose = await get_end_effector_pose(req.args.robot.robot_id, req.args.robot.end_effector)

    if ap.parent:

        obj = glob.SCENE_OBJECT_INSTANCES[ap.parent]
        new_pose = hlp.make_pose_rel(obj.pose, new_pose)

    ori = glob.PROJECT.orientation(req.args.orientation_id)
    ori.orientation = new_pose.orientation

    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.OrientationChanged(events.EventType.UPDATE, data=ori)))
    return None


@scene_needed
@project_needed
async def remove_action_point_orientation_cb(req: rpc.project.RemoveActionPointOrientationRequest, ui: WsClient) -> \
        None:
    """
    Removes orientation.
    :param req:
    :return:
    """

    assert glob.SCENE and glob.PROJECT

    ap, orientation = glob.PROJECT.ap_and_orientation(req.args.action_point_id)

    for act in glob.PROJECT.actions():
        for param in act.parameters:
            if PARAM_PLUGINS[param.type].uses_orientation(glob.PROJECT, act.id, param.id, req.args.orientation_id):
                raise Arcor2Exception(f"Orientation used in action {act.name} (parameter {param.id}).")

    if req.dry_run:
        return None

    ap.orientations = [ori for ori in ap.orientations if ori.id != req.args.orientation_id]

    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.OrientationChanged(events.EventType.REMOVE, data=orientation)))
    return None


async def open_project_cb(req: rpc.project.OpenProjectRequest, ui: WsClient) -> None:

    if glob.PACKAGE_STATE.state != common.PackageStateEnum.STOPPED:
        raise Arcor2Exception("Can't open project while package runs.")

    # TODO validate using project_problems?
    await open_project(req.args.id)

    assert glob.SCENE
    assert glob.PROJECT

    asyncio.ensure_future(notif.broadcast_event(events.OpenProject(data=events.OpenProjectData(glob.SCENE,
                                                                                               glob.PROJECT))))

    return None


@scene_needed
@project_needed
async def save_project_cb(req: rpc.project.SaveProjectRequest, ui: WsClient) -> None:

    assert glob.SCENE and glob.PROJECT
    await storage.update_project(glob.PROJECT)
    glob.PROJECT.modified = (await storage.get_project(glob.PROJECT.id)).modified
    asyncio.ensure_future(notif.broadcast_event(events.ProjectSaved()))
    return None


@no_project
async def new_project_cb(req: rpc.project.NewProjectRequest, ui: WsClient) -> None:

    if glob.PACKAGE_STATE.state != common.PackageStateEnum.STOPPED:
        raise Arcor2Exception("Can't create project while package runs.")

    unique_name(req.args.name, (await project_names()))

    if req.dry_run:
        return None

    if glob.SCENE:
        if glob.SCENE.id != req.args.scene_id:
            raise Arcor2Exception("Another scene is opened.")

        if glob.SCENE.has_changes():
            await storage.update_scene(glob.SCENE)
            glob.SCENE.modified = (await storage.get_scene(glob.SCENE.id)).modified

    else:

        if req.args.scene_id not in {scene.id for scene in (await storage.get_scenes()).items}:
            raise Arcor2Exception("Unknown scene id.")

        await open_scene(req.args.scene_id)

    glob.PROJECT = common.Project(common.uid(), req.args.name, req.args.scene_id, desc=req.args.desc,
                                  has_logic=req.args.has_logic)

    assert glob.SCENE

    asyncio.ensure_future(notif.broadcast_event(events.OpenProject(data=events.OpenProjectData(glob.SCENE,
                                                                                               glob.PROJECT))))
    return None


@scene_needed
@project_needed
async def close_project_cb(req: rpc.project.CloseProjectRequest, ui: WsClient) -> None:

    assert glob.PROJECT

    if not req.args.force and glob.PROJECT.has_changes():
        raise Arcor2Exception("Project has unsaved changes.")

    if req.dry_run:
        return None

    glob.PROJECT = None
    await clear_scene()
    asyncio.ensure_future(notif.broadcast_event(events.ProjectClosed()))

    return None


@scene_needed
@project_needed
async def add_action_point_cb(req: rpc.project.AddActionPointRequest, ui: WsClient) -> None:

    assert glob.PROJECT

    unique_name(req.args.name, glob.PROJECT.action_points_names)

    if not hlp.is_valid_identifier(req.args.name):
        raise Arcor2Exception("Name has to be valid Python identifier.")

    if req.dry_run:
        return None

    ap = common.ProjectActionPoint(common.uid(), req.args.name, req.args.position, req.args.parent)
    glob.PROJECT.action_points.append(ap)

    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.ActionPointChanged(events.EventType.ADD, data=ap)))
    return None


@scene_needed
@project_needed
async def remove_action_point_cb(req: rpc.project.RemoveActionPointRequest, ui: WsClient) -> None:

    assert glob.PROJECT

    ap = glob.PROJECT.action_point(req.args.id)

    for act in glob.PROJECT.actions():
        for param in act.parameters:

            for joints in ap.robot_joints:
                if PARAM_PLUGINS[param.type].uses_robot_joints(glob.PROJECT, act.id, param.id, joints.id):
                    raise Arcor2Exception(f"Joints {joints.name} used in action {act.name} (parameter {param.id}).")

            for ori in ap.orientations:
                if PARAM_PLUGINS[param.type].uses_orientation(glob.PROJECT, act.id, param.id, ori.id):
                    raise Arcor2Exception(f"Orientation {ori.name} used in action {act.name} (parameter {param.id}).")

            # TODO some hypothetical parameter type could use just bare ActionPoint (its position)

    if req.dry_run:
        return None

    glob.PROJECT.action_points = [acp for acp in glob.PROJECT.action_points if acp.id != req.args.id]
    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.ActionPointChanged(events.EventType.REMOVE, data=ap.bare())))
    return None


def check_action_params(project: common.Project, action: common.Action, object_action: object_type.ObjectAction) \
        -> None:

    assert glob.SCENE

    _, action_type = action.parse_type()

    assert action_type == object_action.name

    if len(object_action.parameters) != len(action.parameters):
        raise Arcor2Exception("Unexpected number of parameters.")

    for req_param in object_action.parameters:
        for given_param in action.parameters:
            if req_param.name == given_param.id and req_param.type == given_param.type:
                break
        else:
            raise Arcor2Exception(f"Action parameter {req_param.name}/{req_param.type} is not set or has invalid type.")

    # check values of parameters
    for param in action.parameters:

        if param.type not in PARAM_PLUGINS:
            raise Arcor2Exception(f"Parameter {param.id} of action {action.name} has unknown type: {param.type}.")

        try:
            PARAM_PLUGINS[param.type].value(glob.TYPE_DEF_DICT, glob.SCENE, project, action.id, param.id)
        except ParameterPluginException as e:
            raise Arcor2Exception(f"Parameter {param.id} of action {action.name} has invalid value. {str(e)}")


@scene_needed
@project_needed
async def add_action_cb(req: rpc.project.AddActionRequest, ui: WsClient) -> None:

    assert glob.PROJECT
    assert glob.SCENE

    ap = glob.PROJECT.action_point(req.args.action_point_id)

    unique_name(req.args.name, glob.PROJECT.action_user_names())

    if not hlp.is_valid_identifier(req.args.name):
        raise Arcor2Exception("Action name has to be valid Python identifier.")

    new_action = common.Action(common.uid(), req.args.name, req.args.type, req.args.parameters)

    updated_project = copy.deepcopy(glob.PROJECT)
    updated_ap = updated_project.action_point(req.args.action_point_id)
    updated_ap.actions.append(new_action)

    check_action_params(updated_project, new_action, find_object_action(new_action))

    if req.dry_run:
        return None

    ap.actions.append(new_action)

    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.ActionChanged(events.EventType.ADD, ap.id, data=new_action)))
    return None


@scene_needed
@project_needed
async def update_action_cb(req: rpc.project.UpdateActionRequest, ui: WsClient) -> None:

    assert glob.PROJECT
    assert glob.SCENE

    updated_project = copy.deepcopy(glob.PROJECT)

    updated_action = updated_project.action(req.args.action_id)
    updated_action.parameters = req.args.parameters

    check_action_params(updated_project, updated_action, find_object_action(updated_action))

    if req.dry_run:
        return None

    orig_action = glob.PROJECT.action(req.args.action_id)
    orig_action.parameters = updated_action.parameters

    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.ActionChanged(events.EventType.UPDATE,
                                                                     data=updated_action)))
    return None


def check_action_usage(action: common.Action) -> None:

    assert glob.PROJECT

    for ap in glob.PROJECT.action_points:
        for act in ap.actions:
            for inp in act.inputs:
                if inp.default == action.id:
                    raise Arcor2Exception(f"Action used as an input for another action ({act.name}).")
            for out in act.outputs:
                if out.default == action.id:
                    raise Arcor2Exception(f"Action used as an output for another action ({act.name}).")


@scene_needed
@project_needed
async def remove_action_cb(req: rpc.project.RemoveActionRequest, ui: WsClient) -> None:

    assert glob.PROJECT
    assert glob.SCENE

    ap, action = glob.PROJECT.action_point_and_action(req.args.id)
    check_action_usage(action)

    if req.dry_run:
        return None

    ap.actions = [act for act in ap.actions if act.id != req.args.id]

    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.ActionChanged(events.EventType.REMOVE, data=action.bare())))
    return None


@scene_needed
@project_needed
async def update_action_logic_cb(req: rpc.project.UpdateActionLogicRequest, ui: WsClient) -> None:

    assert glob.PROJECT
    assert glob.SCENE

    action = glob.PROJECT.action(req.args.action_id)

    allowed_values = glob.PROJECT.action_ids() | common.ActionIOEnum.set() | {""}

    for inp in req.args.inputs:
        if inp.default not in allowed_values:
            raise Arcor2Exception("Unknown input value.")

    for out in req.args.outputs:
        if out.default not in allowed_values:
            raise Arcor2Exception("Unknown output value.")

    action.inputs = req.args.inputs
    action.outputs = req.args.outputs

    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.ActionChanged(events.EventType.UPDATE, data=action)))
    return None


@no_project
async def delete_project_cb(req: rpc.project.DeleteProjectRequest, ui: WsClient) -> None:

    project = await storage.get_project(req.args.id)
    await storage.delete_project(req.args.id)
    asyncio.ensure_future(notif.broadcast_event(events.ProjectChanged(events.EventType.REMOVE,
                                                                      data=project.bare())))
    return None


async def rename_project_cb(req: rpc.project.RenameProjectRequest, ui: WsClient) -> None:

    unique_name(req.args.new_name, (await project_names()))

    if req.dry_run:
        return None

    async with managed_project(req.args.project_id) as project:

        project.name = req.args.new_name
        project.update_modified()

        asyncio.ensure_future(notif.broadcast_event(events.ProjectChanged(events.EventType.UPDATE_BASE,
                                                                          data=project.bare())))
    return None


async def copy_project_cb(req: rpc.project.CopyProjectRequest, ui: WsClient) -> None:

    unique_name(req.args.target_name, (await project_names()))

    if req.dry_run:
        return None

    async with managed_project(req.args.source_id, make_copy=True) as project:

        project.name = req.args.target_name
        asyncio.ensure_future(notif.broadcast_event(events.ProjectChanged(events.EventType.UPDATE_BASE,
                                                                          data=project.bare())))

    return None


async def update_project_description_cb(req: rpc.project.UpdateProjectDescriptionRequest, ui: WsClient) -> None:

    async with managed_project(req.args.project_id) as project:

        project.desc = req.args.new_description
        project.update_modified()

        asyncio.ensure_future(notif.broadcast_event(events.ProjectChanged(events.EventType.UPDATE_BASE,
                                                                          data=project.bare())))
    return None


async def update_project_has_logic_cb(req: rpc.project.UpdateProjectHasLogicRequest, ui: WsClient) -> None:

    async with managed_project(req.args.project_id) as project:

        if project.has_logic and not req.args.new_has_logic:

            for act in project.actions():
                act.inputs.clear()
                act.outputs.clear()

                if glob.PROJECT and glob.PROJECT.id == req.args.project_id:
                    asyncio.ensure_future(notif.broadcast_event(events.ActionChanged(events.EventType.UPDATE,
                                                                                     data=act)))

        project.has_logic = req.args.new_has_logic
        project.update_modified()
        asyncio.ensure_future(notif.broadcast_event(events.ProjectChanged(events.EventType.UPDATE_BASE,
                                                                          data=project.bare())))
    return None


@project_needed
async def rename_action_point_joints_cb(req: rpc.project.RenameActionPointJointsRequest, ui: WsClient) -> None:

    assert glob.PROJECT

    ap, joints = glob.PROJECT.ap_and_joints(req.args.joints_id)
    unique_name(req.args.new_name, ap.joints_names())

    if req.dry_run:
        return None

    joints.name = req.args.new_name
    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.JointsChanged(events.EventType.UPDATE_BASE, data=joints)))

    return None


@project_needed
async def rename_action_point_orientation_cb(req: rpc.project.RenameActionPointOrientationRequest, ui: WsClient) -> \
        None:

    assert glob.PROJECT

    ap, ori = glob.PROJECT.ap_and_orientation(req.args.orientation_id)
    unique_name(req.args.new_name, ap.orientation_names())

    if req.dry_run:
        return None

    ori.name = req.args.new_name
    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.OrientationChanged(events.EventType.UPDATE_BASE, data=ori)))

    return None


@project_needed
async def rename_action_cb(req: rpc.project.RenameActionRequest, ui: WsClient) -> None:

    assert glob.PROJECT

    unique_name(req.args.new_name, glob.PROJECT.action_user_names())

    if req.dry_run:
        return None

    act = glob.PROJECT.action(req.args.action_id)
    act.name = req.args.new_name

    glob.PROJECT.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.ActionChanged(events.EventType.UPDATE_BASE, data=act)))

    return None



from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, List, Dict, Union, Set

from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.data import DataException
from arcor2.data.common import ActionMetadata, Pose, Position


class Model3dType(Enum):

    NONE: str = "None"
    BOX: str = "Box"
    CYLINDER: str = "Cylinder"
    SPHERE: str = "Sphere"
    MESH: str = "Mesh"


@dataclass
class MetaModel3d(JsonSchemaMixin):

    id: str
    type: Model3dType

    def __post_init__(self) -> None:

        if isinstance(self.type, str):
            self.type = Model3dType[self.type.upper()]


@dataclass
class Model(JsonSchemaMixin):

    id: str

    @classmethod
    def type(cls) -> Model3dType:
        return Model3dType[cls.__name__.upper()]

    def metamodel(self) -> MetaModel3d:
        return MetaModel3d(self.id, self.type())


@dataclass
class Box(Model):

    size_x: float
    size_y: float
    size_z: float


@dataclass
class Cylinder(Model):

    radius: float
    height: float


@dataclass
class Sphere(Model):
    radius: float


@dataclass
class Mesh(Model):

    uri: str
    focus_points: Optional[List[Pose]] = None


MeshList = List[Mesh]

Models = Union[Box, Sphere, Cylinder, Mesh]

MODEL_MAPPING = {Box.type(): Box,
                 Sphere.type(): Sphere,
                 Cylinder.type(): Cylinder,
                 Mesh.type(): Mesh}


@dataclass
class ObjectType(JsonSchemaMixin):
    """
    Object type, as it is stored in DB.
    """

    id: str
    source: str
    desc: Optional[str] = ""
    model: Optional[MetaModel3d] = None

    def __post_init__(self) -> None:  # TODO workaround for bug (?) in Storage

        if self.model and self.model.type == Model3dType.NONE:
            self.model = None


@dataclass
class ObjectModel(JsonSchemaMixin):

    type: Model3dType
    box: Optional[Box] = None
    cylinder: Optional[Cylinder] = None
    sphere: Optional[Sphere] = None
    mesh: Optional[Mesh] = None

    def model(self) -> Models:
        return getattr(self, str(self.type.value).lower())

    def __post_init__(self) -> None:

        models_list = [self.box, self.cylinder, self.sphere, self.mesh]

        if models_list.count(None) != len(models_list)-1:
            raise DataException("No model specified!")


@dataclass
class ObjectTypeMeta(JsonSchemaMixin):
    """
    Metadata about object type, as it is used in server API.
    """

    type: str
    description: str = field(default_factory=str)
    built_in: bool = False
    base: str = field(default_factory=str)
    object_model: Optional[ObjectModel] = None
    needs_services: Set[str] = field(default_factory=set)
    abstract: bool = False
    disabled: bool = False
    problem: Optional[str] = None

    def to_object_type(self) -> ObjectType:

        ot = ObjectType(self.type, "", self.description)

        if self.object_model:

            if self.object_model.type == Model3dType.MESH:
                assert self.object_model.mesh
                m_id = self.object_model.mesh.id
            else:
                m_id = self.type

            ot.model = MetaModel3d(m_id, self.object_model.type)

        return ot


@dataclass
class ActionParameterMeta(JsonSchemaMixin):

    name: str
    type: str
    dynamic_value: bool = False  # client should ask for allowed values using RPC
    dynamic_value_parents: Optional[Set[str]] = None
    description: Optional[str] = None
    default_value: Optional[str] = None
    extra: Optional[str] = None


@dataclass
class ObjectAction(JsonSchemaMixin):

    name: str
    description: Optional[str] = None
    parameters: List[ActionParameterMeta] = field(default_factory=list)
    meta: ActionMetadata = field(default_factory=ActionMetadata)
    origins: Optional[str] = None
    returns: Optional[str] = None
    disabled: bool = False
    problem: Optional[str] = None


ObjectTypeMetaDict = Dict[str, ObjectTypeMeta]
ObjectActions = List[ObjectAction]
ObjectActionsDict = Dict[str, ObjectActions]


@dataclass
class MeshFocusAction(JsonSchemaMixin):

    mesh_focus_points: List[Position]
    robot_space_points: List[Position]

    def __post_init__(self) -> None:

        assert len(self.mesh_focus_points) == len(self.robot_space_points)


@dataclass
class CollisionModels(JsonSchemaMixin):

    boxes: List[Box] = field(default_factory=list)
    spheres: List[Sphere] = field(default_factory=list)
    cylinders: List[Cylinder] = field(default_factory=list)
    meshes: List[Mesh] = field(default_factory=list)



import socket
import asyncio
from typing import Union, Text, Tuple

from arcor2.server.globals import PORT
from arcor2.data.common import BroadcastInfo

BROADCAST_PORT: int = 6006


def get_ip() -> str:

    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        ip = s.getsockname()[0]
    except socket.error:
        ip = '127.0.0.1'
    finally:
        s.close()
    return ip


def get_broadcast() -> str:

    ip = get_ip()
    ip_arr = ip.split(".")
    ip_arr[-1] = "255"
    return ".".join(ip_arr)


Address = Tuple[str, int]


class BroadcastProtocol(asyncio.DatagramProtocol):

    def __init__(self, target: Address, *, loop: asyncio.AbstractEventLoop = None):
        self.target = target
        self.loop = asyncio.get_event_loop() if loop is None else loop

    def connection_made(self, transport):
        self.transport = transport
        sock: socket.socket = transport.get_extra_info("socket")
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        self.broadcast()

    def datagram_received(self, data: Union[bytes, Text], addr: Address):
        pass

    def broadcast(self):
        self.transport.sendto(BroadcastInfo(socket.gethostname(), PORT).to_json().encode(), self.target)
        self.loop.call_later(1, self.broadcast)


def main():

    loop = asyncio.get_event_loop()

    broadcast_coro = loop.create_datagram_endpoint(lambda: BroadcastProtocol((get_broadcast(), BROADCAST_PORT),
                                                                             loop=loop),
                                                   local_addr=('0.0.0.0', BROADCAST_PORT))

    loop.run_until_complete(broadcast_coro)
    loop.run_forever()
    loop.close()


if __name__ == "__main__":
    main()



from typing import List, Union, Type, Optional, Set
import os

from arcor2.data.common import Pose, Joint
from arcor2.data.robot import RobotMeta
from arcor2.exceptions import Arcor2Exception
import arcor2.helpers as hlp
from arcor2.object_types import Robot, Generic
from arcor2.services import RobotService

from arcor2.server import objects_services_actions as osa, globals as glob


class RobotPoseException(Arcor2Exception):
    pass


async def collision(obj: Generic,
                    rs: Optional[RobotService] = None, *, add: bool = False, remove: bool = False) -> None:
    """

    :param obj: Instance of the object.
    :param add:
    :param remove:
    :param rs:
    :return:
    """

    assert add ^ remove

    if not obj.collision_model:
        return

    if rs is None:
        rs = osa.find_robot_service()
    if rs:
        try:
            # TODO notify user somehow when something went wrong?
            await hlp.run_in_executor(rs.add_collision if add else rs.remove_collision, obj)
        except Arcor2Exception as e:
            await glob.logger.error(e)


async def get_end_effectors(robot_id: str) -> Set[str]:
    """
    :param robot_id:
    :return: IDs of existing end effectors.
    """

    robot_inst = await osa.get_robot_instance(robot_id)

    if isinstance(robot_inst, Robot):
        return await hlp.run_in_executor(robot_inst.get_end_effectors_ids)
    else:
        return await hlp.run_in_executor(robot_inst.get_end_effectors_ids, robot_id)


async def get_grippers(robot_id: str) -> Set[str]:
    """
    :param robot_id:
    :return: IDs of existing grippers.
    """

    robot_inst = await osa.get_robot_instance(robot_id)

    if isinstance(robot_inst, Robot):
        return await hlp.run_in_executor(robot_inst.grippers)
    else:
        return await hlp.run_in_executor(robot_inst.grippers, robot_id)


async def get_suctions(robot_id: str) -> Set[str]:
    """
    :param robot_id:
    :return: IDs of existing suctions.
    """

    robot_inst = await osa.get_robot_instance(robot_id)

    if isinstance(robot_inst, Robot):
        return await hlp.run_in_executor(robot_inst.suctions)
    else:
        return await hlp.run_in_executor(robot_inst.suctions, robot_id)


async def get_end_effector_pose(robot_id: str, end_effector: str) -> Pose:
    """
    :param robot_id:
    :param end_effector:
    :return: Global pose
    """

    robot_inst = await osa.get_robot_instance(robot_id, end_effector)

    if isinstance(robot_inst, Robot):
        return await hlp.run_in_executor(robot_inst.get_end_effector_pose, end_effector)
    else:
        return await hlp.run_in_executor(robot_inst.get_end_effector_pose, robot_id, end_effector)


async def get_robot_joints(robot_id: str) -> List[Joint]:
    """
    :param robot_id:
    :return: List of joints
    """

    robot_inst = await osa.get_robot_instance(robot_id)

    if isinstance(robot_inst, Robot):
        return await hlp.run_in_executor(robot_inst.robot_joints)
    else:
        return await hlp.run_in_executor(robot_inst.robot_joints, robot_id)


async def get_robot_meta(robot_type: Union[Type[Robot], Type[RobotService]]) -> None:

    meta = RobotMeta(robot_type.__name__)
    meta.features.focus = hasattr(robot_type, "focus")  # TODO more sophisticated test? (attr(s) and return value?)

    if issubclass(robot_type, Robot) and robot_type.urdf_package_path:
        meta.urdf_package_filename = os.path.split(robot_type.urdf_package_path)[1]

    glob.ROBOT_META[robot_type.__name__] = meta



# -*- coding: utf-8 -*-

from typing import List, Optional
from datetime import datetime

from dataclasses import dataclass, field
from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.data.common import PackageState, ActionState, CurrentAction
from arcor2.data.execution import PackageMeta
from arcor2.data.rpc.common import IdArgs, Request, Response, wo_suffix


@dataclass
class BuildProjectArgs(JsonSchemaMixin):

    project_id: str
    package_name: str


@dataclass
class BuildProjectRequest(Request):
    """
    Calls Build service to generate execution package and uploads it to the Execution service.
    """

    args: BuildProjectArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class BuildProjectData(JsonSchemaMixin):

    package_id: str


@dataclass
class BuildProjectResponse(Response):

    data: Optional[BuildProjectData] = None
    response: str = field(default=BuildProjectRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class UploadPackageArgs(JsonSchemaMixin):

    id: str = field(metadata=dict(description="Id of the execution package."))
    data: str = field(metadata=dict(description="Base64 encoded content of the zip file."))


@dataclass
class UploadPackageRequest(Request):

    args: UploadPackageArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class UploadPackageResponse(Response):

    response: str = field(default=UploadPackageRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class ListPackagesRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ModifiedFile(JsonSchemaMixin):

    filename: str
    modified: datetime


@dataclass
class PackageSummary(JsonSchemaMixin):

    id: str
    project_id: str
    modified: datetime = field(metadata=dict(
        description="Last modification of the project embedded in the execution package."))
    # modified_files: List[ModifiedFile] = field(default_factory=list)
    package_meta: PackageMeta = field(metadata=dict(description="Content of 'package.json'."))


@dataclass
class ListPackagesResponse(Response):

    data: List[PackageSummary] = field(default_factory=list)
    response: str = field(default=ListPackagesRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------

@dataclass
class DeletePackageRequest(Request):

    args: IdArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class DeletePackageResponse(Response):

    response: str = field(default=DeletePackageRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class RenamePackageArgs(JsonSchemaMixin):

    package_id: str
    new_name: str


@dataclass
class RenamePackageRequest(Request):

    args: RenamePackageArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RenamePackageResponse(Response):

    response: str = field(default=RenamePackageRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class RunPackageRequest(Request):

    args: IdArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RunPackageResponse(Response):

    response: str = field(default=RunPackageRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class StopPackageRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class StopPackageResponse(Response):

    response: str = field(default=StopPackageRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class PausePackageRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class PausePackageResponse(Response):

    response: str = field(default=PausePackageRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------

@dataclass
class PackageStateRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class PackageStateData(JsonSchemaMixin):

    project: PackageState = field(default_factory=PackageState)
    action: Optional[ActionState] = None
    action_args: Optional[CurrentAction] = None


@dataclass
class PackageStateResponse(Response):

    data: PackageStateData = field(default_factory=PackageStateData)
    response: str = field(default=PackageStateRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class ResumePackageRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ResumePackageResponse(Response):

    response: str = field(default=ResumePackageRequest.request, init=False)



from typing import List

from arcor2.data.common import Project, Scene
from arcor2.parameter_plugins.base import ParameterPlugin, TypesDict
from arcor2.parameter_plugins.list import ListParameterPlugin, get_type_name


class StringPlugin(ParameterPlugin):

    @classmethod
    def type(cls):
        return str

    @classmethod
    def type_name(cls) -> str:
        return "string"

    @classmethod
    def value(cls, type_defs: TypesDict, scene: Scene, project: Project, action_id: str, parameter_id: str) -> str:
        return cls.type()(super(StringPlugin, cls).value(type_defs, scene, project, action_id, parameter_id))


class StringListPlugin(ListParameterPlugin):

    @classmethod
    def type(cls):
        return List[str]

    @classmethod
    def type_name(cls) -> str:
        return get_type_name(StringPlugin)

    @classmethod
    def value(cls, type_defs: TypesDict, scene: Scene, project: Project, action_id: str, parameter_id: str) \
            -> List[str]:
        return super(StringListPlugin, cls).value(type_defs, scene, project, action_id, parameter_id)



import json
import requests
from typing import Type, TypeVar, Dict, Callable, List, Union, Any, Optional, Sequence
import io
import functools

from dataclasses_jsonschema import ValidationError, JsonSchemaMixin
from PIL import Image, UnidentifiedImageError  # type: ignore

from arcor2.exceptions import Arcor2Exception
from arcor2.helpers import camel_case_to_snake_case, snake_case_to_camel_case


class RestException(Arcor2Exception):
    pass


TIMEOUT = (1.0, 20.0)  # connect, read

HEADERS = {'accept': 'application/json', 'content-type': 'application/json'}

T = TypeVar('T', bound=JsonSchemaMixin)
S = TypeVar('S', str, int, float, bool)

OptionalData = Optional[Union[JsonSchemaMixin, Sequence[JsonSchemaMixin]]]
ParamsDict = Optional[Dict[str, Any]]

SESSION = requests.session()


def handle_exceptions(exception_type: Type[Arcor2Exception] = Arcor2Exception, message: Optional[str] = None):
    def _handle_exceptions(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):

            try:
                return func(*args, **kwargs)
            except RestException as e:
                if message is not None:
                    raise exception_type(message, str(e)) from e
                else:
                    raise exception_type(e.message) from e

        return wrapper
    return _handle_exceptions


def convert_keys(d: Union[Dict, List], func: Callable[[str], str]) -> Union[Dict, List]:

    if isinstance(d, dict):
        new_dict = {}
        for k, v in d.items():
            new_dict[func(k)] = convert_keys(v, func)
        return new_dict
    elif isinstance(d, list):
        new_list: List[Any] = []
        for dd in d:
            new_list.append(convert_keys(dd, func))
        return new_list

    return d


def handle_response(resp: requests.Response) -> None:
    try:
        resp.raise_for_status()
    except requests.exceptions.RequestException as e:

        try:
            resp_body = json.loads(resp.content)
        except json.JSONDecodeError:
            raise RestException(resp.content.decode("utf-8"), str(e)) from e

        try:
            raise RestException(resp_body["message"], str(e)) from e
        except (KeyError, TypeError):  # TypeError is for case when resp_body is just string
            raise RestException(str(resp_body), str(e)) from e


def _send(url: str, op: Callable, data: OptionalData = None,
          params: ParamsDict = None, get_response=False) -> Union[None, Dict, List]:

    if data:
        if isinstance(data, list):
            d = []
            for dd in data:
                d.append(convert_keys(dd.to_dict(), snake_case_to_camel_case))
        else:
            d = convert_keys(data.to_dict(), snake_case_to_camel_case)  # type: ignore
    else:
        d = {}  # type: ignore

    if params:
        params = convert_keys(params, snake_case_to_camel_case)  # type: ignore
    else:
        params = {}

    try:
        resp = op(url, data=json.dumps(d), timeout=TIMEOUT, headers=HEADERS,
                  params=params)
    except requests.exceptions.RequestException as e:
        raise RestException("Catastrophic system error.", str(e)) from e

    handle_response(resp)

    if not get_response:
        return None

    try:
        return convert_keys(json.loads(resp.text), camel_case_to_snake_case)
    except (json.JSONDecodeError, TypeError) as e:
        raise RestException("Invalid JSON.", str(e)) from e


def post(url: str, data: JsonSchemaMixin, params: ParamsDict = None):
    _send(url, SESSION.post, data, params)


def put_returning_primitive(url: str, desired_type: Type[S], data: OptionalData = None,
                            params: ParamsDict = None) -> S:

    try:
        return desired_type(_send(url, SESSION.put, data, params, get_response=True))  # type: ignore
    except ValueError as e:
        raise RestException(e) from e


def put(url: str, data: OptionalData = None, params: ParamsDict = None, data_cls: Type[T] = None) -> T:
    ret = _send(url, SESSION.put, data, params, get_response=data_cls is not None)  # type: ignore

    if not data_cls:
        return None  # type: ignore

    try:
        return data_cls.from_dict(ret)  # type: ignore
    except ValidationError as e:
        print(f'{data_cls.__name__}: validation error "{e}" while parsing "{data}".')
        raise RestException("Invalid data.", str(e)) from e


def put_returning_list(url: str, data: OptionalData = None,
                       params: ParamsDict = None, data_cls: Type[T] = None) -> List[T]:

    ret = _send(url, SESSION.put, data, params, get_response=data_cls is not None)  # type: ignore

    if not data_cls:
        return []  # type: ignore

    assert isinstance(ret, list)

    d = []
    for dd in ret:
        try:
            d.append(data_cls.from_dict(dd))
        except ValidationError as e:
            print(f'{data_cls.__name__}: validation error "{e}" while parsing "{ret}".')
            raise RestException("Invalid data.", str(e)) from e

    return d


def delete(url: str):

    try:
        resp = SESSION.delete(url, timeout=TIMEOUT, headers=HEADERS)
    except requests.exceptions.RequestException as e:
        raise RestException("Catastrophic system error.", str(e)) from e

    handle_response(resp)


def get_data(url: str, body: Optional[JsonSchemaMixin] = None, params: ParamsDict = None) -> Union[Dict, List]:

    data = _get(url, body, params)

    if not isinstance(data, (list, dict)):
        raise RestException("Invalid data, not list or dict.")

    return convert_keys(data, camel_case_to_snake_case)


def _get_response(url: str, body: Optional[JsonSchemaMixin] = None, params: ParamsDict = None) -> requests.Response:

    if body is None:
        body_dict = {}  # type: ignore
    else:
        body_dict = body.to_dict()

    if params:
        params = convert_keys(params, snake_case_to_camel_case)  # type: ignore
    else:
        params = {}

    try:
        resp = SESSION.get(url, timeout=TIMEOUT, data=body_dict, params=params, allow_redirects=True,
                           headers=HEADERS)
    except requests.exceptions.RequestException as e:
        raise RestException("Catastrophic system error.", str(e)) from e

    handle_response(resp)

    return resp


def _get(url: str, body: Optional[JsonSchemaMixin] = None, params: ParamsDict = None) -> Any:

    resp = _get_response(url, body, params)

    try:
        return json.loads(resp.text)
    except (json.JSONDecodeError, TypeError) as e:
        raise RestException("Invalid JSON.", str(e)) from e


def get_image(url: str) -> Image.Image:

    # TODO check content type?
    try:
        return Image.open(io.BytesIO(_get_response(url).content))
    except (UnidentifiedImageError, TypeError) as e:
        raise RestException("Invalid image.", str(e)) from e


def get_primitive(url: str, desired_type: Type[S], body: Optional[JsonSchemaMixin] = None,
                  params: ParamsDict = None) -> S:

    value = _get(url, body, params)

    try:
        return desired_type(value)
    except ValueError as e:
        raise RestException("Invalid primitive.", str(e)) from e


def get_list(url: str, data_cls: Type[T], body: Optional[JsonSchemaMixin] = None, params: ParamsDict = None) -> List[T]:

    data = get_data(url, body, params)

    ret: List[T] = []

    for val in data:
        try:
            ret.append(data_cls.from_dict(val))
        except ValidationError as e:
            print(f'{data_cls.__name__}: validation error "{e}" while parsing "{data}".')
            raise RestException("Invalid data.", str(e)) from e

    return ret


def get_list_primitive(url: str, desired_type: Type[S], body: Optional[JsonSchemaMixin] = None,
                       params: ParamsDict = None) -> List[S]:

    data = get_data(url, body, params)

    ret: List[S] = []

    for val in data:
        ret.append(desired_type(val))

    return ret


def get(url: str, data_cls: Type[T], body: Optional[JsonSchemaMixin] = None, params: ParamsDict = None) -> T:

    data = get_data(url, body, params)

    assert isinstance(data, dict)

    try:
        return data_cls.from_dict(data)
    except ValidationError as e:
        raise RestException("Invalid data.", str(e)) from e


def download(url: str, path: str, body: Optional[JsonSchemaMixin] = None, params: ParamsDict = None) -> None:
    # TODO check content type

    r = _get_response(url, body, params)
    with open(path, 'wb') as file:
        file.write(r.content)



import inspect
from typing import Dict, Tuple, Type

import arcor2.data.events
from arcor2.data import rpc
from arcor2.data.events import Event
from arcor2.data.rpc.common import Request, Response

RPC_MAPPING: Dict[str, Tuple[Type[Request], Type[Response]]] = {}

_requests: Dict[str, Type[Request]] = {}
_responses: Dict[str, Type[Response]] = {}

for rpc_module_name, rpc_module in inspect.getmembers(rpc, inspect.ismodule):
    for name, obj in inspect.getmembers(rpc_module, inspect.isclass):

        if not inspect.isclass(obj):
            continue

        if issubclass(obj, Request) and obj != Request:
            _requests[obj.request] = obj
        elif issubclass(obj, Response) and obj != Response:
            _responses[obj.response] = obj

assert _requests.keys() == _responses.keys(),\
    f"There is difference between requests/responses: " \
    f"{set(_requests.keys()).symmetric_difference(set(_responses.keys()))}"

for k, v in _requests.items():
    RPC_MAPPING[k] = (v, _responses[k])

EVENT_MAPPING: Dict[str, Type[Event]] = {}

for name, obj in inspect.getmembers(arcor2.data.events):

    if inspect.isclass(obj) and issubclass(obj, Event) and obj != Event:
        EVENT_MAPPING[obj.event] = obj



from typing import Tuple, Dict, Union

from arcor2.data.common import Project, Action, ActionIOEnum
from arcor2.exceptions import Arcor2Exception


class ProjectException(Arcor2Exception):
    pass


def get_actions_cache(project: Project) -> Tuple[Dict[str, Action], Union[str, None], Union[str, None]]:

    actions_cache = {}
    first_action_id = None
    last_action_id = None

    for aps in project.action_points:
        for act in aps.actions:
            actions_cache[act.id] = act
            if act.inputs and act.inputs[0].default == ActionIOEnum.FIRST.value:
                if first_action_id is not None:
                    raise ProjectException("Multiple starts.")
                first_action_id = act.id
            elif act.outputs and act.outputs[0].default == ActionIOEnum.LAST.value:
                if last_action_id is not None:
                    raise ProjectException("Multiple ends.")
                last_action_id = act.id

    if first_action_id not in actions_cache:
        raise ProjectException(f"Unknown start action: {first_action_id}.")

    if last_action_id not in actions_cache:
        raise ProjectException(f"Unknown end action: {last_action_id}.")

    return actions_cache, first_action_id, last_action_id


def clear_project_logic(project: Project) -> None:  # TODO method of Project class?

    for act_point in project.action_points:
        for action in act_point.actions:
            action.inputs.clear()
            action.outputs.clear()



from typing import List
import time

from arcor2.object_types import Generic
from arcor2.action import action
from arcor2.data.common import ActionMetadata


class Tester(Generic):
    """
    A generic tester
    """

    def __init__(self, *args, **kwargs):
        super(Tester, self).__init__(*args, **kwargs)
        self._cancel: bool = False
        self._param1: str = ""

    @action
    def run_test(self, seq: List[float], seq_id: List[str], eqp_res: List[bool]) -> List[bool]:
        return [bool(seq), bool(seq_id), bool(eqp_res)]

    @action
    def long_running_action(self) -> None:
        for _ in range(60):
            time.sleep(1)
            if self._cancel:
                self._cancel = False
                break

    @action
    def long_running_action_with_params(self, param1: str) -> None:
        self._param1 = param1
        for _ in range(60):
            time.sleep(1)
            if self._cancel:
                self._cancel = False
                break

    def simple_cancel(self) -> None:
        self._cancel = True

    def cancel_with_param(self, param1: str) -> None:
        assert param1 == self._param1
        self._cancel = True

    run_test.__action__ = ActionMetadata(blocking=True)
    long_running_action.__action__ = ActionMetadata(blocking=True)
    long_running_action_with_params.__action__ = ActionMetadata(blocking=True)


Tester.CANCEL_MAPPING[Tester.long_running_action.__name__] = Tester.simple_cancel.__name__
Tester.CANCEL_MAPPING[Tester.long_running_action_with_params.__name__] = Tester.cancel_with_param.__name__



import os

from arcor2.data.common import Project, Scene, ProjectSources, IdDescList
from arcor2.data.object_type import ObjectType, Models, MODEL_MAPPING, Model3dType, Mesh, MeshList
from arcor2.data.services import ServiceType
from arcor2.exceptions import Arcor2Exception
from arcor2 import rest

URL = os.getenv("ARCOR2_PERSISTENT_STORAGE_URL", "http://0.0.0.0:11000")

# TODO thread to poll changes? how to "detect" changes?


class PersistentStorageException(Arcor2Exception):
    pass


@rest.handle_exceptions(PersistentStorageException)
def get_mesh(mesh_id) -> Mesh:
    return rest.get(f"{URL}/models/{mesh_id}/mesh", Mesh)


@rest.handle_exceptions(PersistentStorageException)
def get_meshes() -> MeshList:
    return rest.get_list(f"{URL}/models/meshes", Mesh)


@rest.handle_exceptions(PersistentStorageException)
def get_model(model_id: str, model_type: Model3dType) -> Models:
    return rest.get(f"{URL}/models/{model_id}/{model_type.value}", MODEL_MAPPING[model_type])


@rest.handle_exceptions(PersistentStorageException)
def put_model(model: Models) -> None:
    rest.put(f"{URL}/models/{model.__class__.__name__.lower()}", model)


@rest.handle_exceptions(PersistentStorageException)
def get_projects() -> IdDescList:
    return rest.get(f"{URL}/projects", IdDescList)


@rest.handle_exceptions(PersistentStorageException)
def get_scenes() -> IdDescList:
    return rest.get(f"{URL}/scenes", IdDescList)


@rest.handle_exceptions(PersistentStorageException)
def get_project(project_id: str) -> Project:
    return rest.get(f"{URL}/project/{project_id}", Project)


@rest.handle_exceptions(PersistentStorageException)
def get_project_sources(project_id: str) -> ProjectSources:
    return rest.get(f"{URL}/project/{project_id}/sources", ProjectSources)


@rest.handle_exceptions(PersistentStorageException)
def get_scene(scene_id: str) -> Scene:
    return rest.get(f"{URL}/scene/{scene_id}", Scene)


@rest.handle_exceptions(PersistentStorageException)
def get_object_type(object_type_id: str) -> ObjectType:
    return rest.get(f"{URL}/object_types/{object_type_id}", ObjectType)


@rest.handle_exceptions(PersistentStorageException)
def get_service_type(service_type_id: str) -> ServiceType:
    return rest.get(f"{URL}/service_type/{service_type_id}", ServiceType)


@rest.handle_exceptions(PersistentStorageException)
def get_object_type_ids() -> IdDescList:
    return rest.get(f"{URL}/object_types", IdDescList)


@rest.handle_exceptions(PersistentStorageException)
def get_service_type_ids() -> IdDescList:
    return rest.get(f"{URL}/service_types", IdDescList)


@rest.handle_exceptions(PersistentStorageException)
def update_project(project: Project) -> None:

    assert project.id
    rest.put(f"{URL}/project", project)


@rest.handle_exceptions(PersistentStorageException)
def update_scene(scene: Scene) -> None:

    assert scene.id
    rest.put(f"{URL}/scene", scene)


@rest.handle_exceptions(PersistentStorageException)
def update_project_sources(project_sources: ProjectSources) -> None:

    assert project_sources.id
    rest.post(f"{URL}/project/sources", project_sources)


@rest.handle_exceptions(PersistentStorageException)
def update_object_type(object_type: ObjectType) -> None:

    assert object_type.id
    rest.put(f"{URL}/object_type", object_type)


@rest.handle_exceptions(PersistentStorageException)
def update_service_type(service_type: ServiceType) -> None:

    assert service_type.id
    rest.put(f"{URL}/service_type", service_type)


@rest.handle_exceptions(PersistentStorageException)
def delete_scene(scene_id: str) -> None:
    return rest.delete(f"{URL}/scene/{scene_id}")


@rest.handle_exceptions(PersistentStorageException)
def delete_project(project_id: str) -> None:
    return rest.delete(f"{URL}/project/{project_id}")



from arcor2.object_types import Generic
from arcor2.data.common import ActionMetadata
from arcor2.action import action


class Box(Generic):

    @action
    def test(self) -> None:
        pass

    @action
    def action_with_unknown_parameter_type(self, param: ActionMetadata) -> bool:
        return True

    test.__action__ = ActionMetadata()  # type: ignore
    action_with_unknown_parameter_type.__action__ = ActionMetadata()  # type: ignore



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from typing import Dict, Set, Union, Optional, DefaultDict, Any
import os
from collections import defaultdict

from websockets.server import WebSocketServerProtocol as WsClient
from aiologger import Logger  # type: ignore
from aiologger.levels import LogLevel  # type: ignore

from arcor2 import helpers as hlp
from arcor2.data.common import Scene, Project, PackageState, ActionState, CurrentAction
from arcor2.data.execution import PackageInfo
from arcor2.data.object_type import ObjectActionsDict, ObjectTypeMetaDict
from arcor2.data.services import ServiceTypeMetaDict
from arcor2.data.robot import RobotMeta
from arcor2.object_types import Generic
from arcor2.services import Service
from arcor2.parameter_plugins.base import TypesDict
from arcor2.nodes.execution import PORT as EXE_PORT
from arcor2.nodes.build import PORT as BUILD_PORT

logger = Logger.with_default_handlers(name='server', formatter=hlp.aiologger_formatter(), level=LogLevel.DEBUG)
VERBOSE: bool = False

MANAGER_URL = os.getenv("ARCOR2_EXECUTION_URL", f"ws://0.0.0.0:{EXE_PORT}")
BUILDER_URL = os.getenv("ARCOR2_BUILDER_URL", f"http://0.0.0.0:{BUILD_PORT}")

PORT: int = int(os.getenv("ARCOR2_SERVER_PORT", 6789))

SCENE: Union[Scene, None] = None
PROJECT: Union[Project, None] = None

INTERFACES: Set[WsClient] = set()

OBJECT_TYPES: ObjectTypeMetaDict = {}
SERVICE_TYPES: ServiceTypeMetaDict = {}
ROBOT_META: Dict[str, RobotMeta] = {}
TYPE_DEF_DICT: TypesDict = {}

# TODO merge it into one dict?
SCENE_OBJECT_INSTANCES: Dict[str, Generic] = {}
SERVICES_INSTANCES: Dict[str, Service] = {}

ACTIONS: ObjectActionsDict = {}  # used for actions of both object_types / services

RUNNING_ACTION: Optional[str] = None  # ID of an action that is being executed during project editing
RUNNING_ACTION_PARAMS: Optional[Dict[str, Any]] = None

PACKAGE_STATE = PackageState()
PACKAGE_INFO: Optional[PackageInfo] = None
ACTION_STATE: Optional[ActionState] = None
CURRENT_ACTION: Optional[CurrentAction] = None

RegisteredUiDict = DefaultDict[str, Set[WsClient]]

ROBOT_JOINTS_REGISTERED_UIS: RegisteredUiDict = defaultdict(lambda: set())  # robot, UIs
ROBOT_EEF_REGISTERED_UIS: RegisteredUiDict = defaultdict(lambda: set())  # robot, UIs



from typing import Set, Union

from horast import parse
from typed_ast.ast3 import Expr, Pass, Module, Call, Attribute

from arcor2.data.common import Project, Scene, Action, ActionIO, ActionIOEnum
from arcor2.helpers import camel_case_to_snake_case
from arcor2.project_utils import get_actions_cache, ProjectException
from arcor2.source import SCRIPT_HEADER, SourceException
from arcor2.source.object_types import fix_object_name
from arcor2.source.utils import main_loop_body, empty_script_tree, add_import, \
    append_method_call, tree_to_str, get_name_attr, clean
from arcor2.source.object_types import object_instance_from_res
import arcor2.object_types
from arcor2.exceptions import Arcor2Exception


def program_src(project: Project, scene: Scene, built_in_objects: Set[str], add_logic: bool = True) -> str:

    tree = empty_script_tree(add_main_loop=add_logic)

    # get object instances from resources object
    for obj in scene.objects:

        if obj.type in built_in_objects:
            add_import(tree, arcor2.object_types.__name__, obj.type, try_to_import=False)
        else:
            add_import(tree, "object_types." + camel_case_to_snake_case(obj.type), obj.type, try_to_import=False)

        object_instance_from_res(tree, obj.name, obj.id, obj.type, "objects")

    for srv in scene.services:
        add_import(tree, "services." + camel_case_to_snake_case(srv.type), srv.type, try_to_import=False)
        object_instance_from_res(tree, srv.type, srv.type, srv.type, "services")

    if add_logic:
        add_logic_to_loop(tree, scene, project)

    return SCRIPT_HEADER + tree_to_str(tree)


def get_logic_from_source(source_code: str, project: Project) -> None:

    tree = parse(source_code)

    assert isinstance(tree, Module)

    try:
        actions_cache, _, _ = get_actions_cache(project)
    except ProjectException as e:
        raise SourceException(e)
    # objects_cache = get_objects_cache(project, id_to_var=True)

    found_actions: Set[str] = set()

    loop = main_loop_body(tree)

    last_action: Union[None, Action] = None

    for node_idx, node in enumerate(loop):

        # simple checks for expected 'syntax' of action calls (e.g. 'robot.move_to(**res.MoveToBoxIN)')
        if not isinstance(node, Expr) or not isinstance(node.value, Call) or not isinstance(node.value.func, Attribute):
            raise SourceException("Unexpected content.")

        try:
            val = node.value
            obj_id = val.func.value.id  # type: ignore
            method = val.func.attr  # type: ignore
        except (AttributeError, IndexError) as e:
            print(e)
            raise SourceException("Script has unexpected content.")

        """
        Support for both:
            robot.move_to(res.MoveToBoxIN)  # args
        ...as well as
            robot.move_to(**res.MoveToBoxIN)  # kwargs
        """
        if len(val.args) == 1 and not val.keywords:
            action_id = val.args[0].attr  # type: ignore
        elif len(val.keywords) == 1 and not val.args:
            action_id = val.keywords[0].value.attr  # type: ignore
        else:
            raise SourceException("Unexpected argument(s) to the action.")

        if action_id in found_actions:
            raise SourceException(f"Duplicate action: {action_id}.")
        found_actions.add(action_id)

        # TODO test if object instance exists
        # raise GenerateSourceException(f"Unknown object id {obj_id}.")

        try:
            action = actions_cache[action_id]
        except KeyError:
            raise SourceException(f"Unknown action {action_id}.")

        at_obj, at_method = action.type.split("/")
        at_obj = camel_case_to_snake_case(at_obj)  # convert obj id into script variable name

        if at_obj != obj_id or at_method != method:
            raise SourceException(f"Action type {action.type} does not correspond to source, where it is"
                                  f" {obj_id}/{method}.")

        action.inputs.clear()
        action.outputs.clear()

        if node_idx == 0:
            action.inputs.append(ActionIO(ActionIOEnum.FIRST.value))
        else:
            assert last_action is not None
            action.inputs.append(ActionIO(last_action.id))

        if node_idx > 0:
            assert last_action is not None
            actions_cache[last_action.id].outputs.append(ActionIO(action.id))

        if node_idx == len(loop)-1:
            action.outputs.append(ActionIO(ActionIOEnum.LAST.value))

        last_action = action


def add_logic_to_loop(tree: Module, scene: Scene, project: Project) -> None:

    loop = main_loop_body(tree)

    try:
        actions_cache, first_action_id, last_action_id = get_actions_cache(project)
    except ProjectException as e:
        raise SourceException(e)

    if first_action_id is None:
        raise SourceException("'start' action not found.")

    if last_action_id is None:
        raise SourceException("'end' action not found.")

    next_action_id = first_action_id

    while True:

        act = actions_cache[next_action_id]

        if len(loop) == 1 and isinstance(loop[0], Pass):
            # pass is not necessary now
            loop.clear()

        ac_obj, ac_type = act.type.split('/')

        # for scene objects, convert ID to name
        try:
            ac_obj = scene.object(ac_obj).name
        except Arcor2Exception:
            pass

        append_method_call(loop, fix_object_name(ac_obj), ac_type, [get_name_attr("res", clean(act.name))], [])

        if act.id == last_action_id:
            break

        next_action_id = act.outputs[0].default



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import Callable, Any, Union, Dict, Optional, Type
import asyncio
import shutil

from arcor2 import service_types_utils as stu, object_types_utils as otu
from arcor2.data.object_type import ObjectModel, ObjectTypeMeta, ObjectTypeMetaDict, ObjectActionsDict
from arcor2.data.services import ServiceTypeMeta, ServiceTypeMetaDict
from arcor2 import aio_persistent_storage as storage
from arcor2.object_types import Robot
from arcor2.exceptions import Arcor2Exception
from arcor2.services import RobotService, Service
from arcor2.object_types import Generic
from arcor2.parameter_plugins import TYPE_TO_PLUGIN
import arcor2.helpers as hlp
from arcor2.data import events

from arcor2.server import globals as glob, settings
from arcor2.server.robot import get_robot_meta
from arcor2.server import notifications as notif


def find_robot_service() -> Union[None, RobotService]:

    for srv in glob.SERVICES_INSTANCES.values():
        if isinstance(srv, RobotService):
            return srv
    else:
        return None


def get_obj_type_name(object_id: str) -> str:

    return glob.SCENE_OBJECT_INSTANCES[object_id].__class__.__name__


def valid_object_types() -> ObjectTypeMetaDict:
    """
    To get only valid (not disabled) types.
    :return:
    """

    return {obj_type: obj for obj_type, obj in glob.OBJECT_TYPES.items() if not obj.disabled}


def valid_service_types() -> ServiceTypeMetaDict:
    """
    To get only valid (not disabled) types.
    :return:
    """

    return {srv_type: srv for srv_type, srv in glob.SERVICE_TYPES.items() if not srv.disabled}


async def get_service_types() -> None:

    service_types: ServiceTypeMetaDict = {}

    srv_ids = await storage.get_service_type_ids()

    # TODO do it in parallel
    for srv_id in srv_ids.items:

        srv_type = await storage.get_service_type(srv_id.id)

        try:
            type_def = hlp.type_def_from_source(srv_type.source, srv_type.id, Service)
            meta = stu.meta_from_def(type_def)
            service_types[srv_id.id] = meta
        except Arcor2Exception as e:
            await glob.logger.warning(f"Disabling service type {srv_type.id}.")
            await glob.logger.debug(e, exc_info=True)
            service_types[srv_id.id] = ServiceTypeMeta(srv_id.id, "Service not available.", disabled=True,
                                                       problem=e.message)
            continue

        if not meta.configuration_ids:
            meta.disabled = True
            meta.problem = "No configuration available."
            continue

        glob.TYPE_DEF_DICT[srv_id.id] = type_def

        if issubclass(type_def, RobotService):
            asyncio.ensure_future(get_robot_meta(type_def))

    glob.SERVICE_TYPES = service_types


def handle_robot_urdf(robot: Type[Robot]) -> None:

    if not robot.urdf_package_path:
        return

    shutil.copy(robot.urdf_package_path, settings.URDF_PATH)


async def get_object_types() -> None:

    object_types: ObjectTypeMetaDict = otu.built_in_types_meta()

    obj_ids = await storage.get_object_type_ids()

    # TODO do it in parallel
    for obj_id in obj_ids.items:
        obj = await storage.get_object_type(obj_id.id)
        try:
            type_def = hlp.type_def_from_source(obj.source, obj.id, Generic)
            meta = otu.meta_from_def(type_def)
            object_types[obj.id] = meta
        except (otu.ObjectTypeException, hlp.TypeDefException) as e:
            await glob.logger.warning(f"Disabling object type {obj.id}.")
            await glob.logger.debug(e, exc_info=True)
            object_types[obj.id] = ObjectTypeMeta(obj_id.id, "Object type disabled.", disabled=True,
                                                  problem=e.message)
            continue

        for srv in meta.needs_services:
            try:
                if glob.SERVICE_TYPES[srv].disabled:
                    meta.disabled = True
                    meta.problem = f"Depends on disabled service '{srv}'."
                    break
            except KeyError:
                meta.disabled = True
                meta.problem = f"Depends on unknown service '{srv}'."
                break

        glob.TYPE_DEF_DICT[obj.id] = type_def

        if obj.model:
            model = await storage.get_model(obj.model.id, obj.model.type)
            kwargs = {model.type().value.lower(): model}
            object_types[obj.id].object_model = ObjectModel(model.type(), **kwargs)  # type: ignore

        if issubclass(type_def, Robot):
            asyncio.ensure_future(get_robot_meta(type_def))
            asyncio.ensure_future(hlp.run_in_executor(handle_robot_urdf, type_def))

    # if description is missing, try to get it from ancestor(s)
    for obj_type, obj_meta in object_types.items():
        if obj_meta.description:
            continue

        try:
            obj_meta.description = otu.obj_description_from_base(object_types, obj_meta)
        except otu.DataError as e:
            await glob.logger.error(f"Failed to get info from base for {obj_type}, error: '{e}'.")

    glob.OBJECT_TYPES = object_types


async def get_object_actions() -> None:  # TODO do it in parallel

    global ACTIONS

    object_actions_dict: ObjectActionsDict = otu.built_in_types_actions(TYPE_TO_PLUGIN)

    valid_types = valid_object_types()

    for obj_type, obj in valid_types.items():

        if obj.built_in:  # built-in types are already there
            continue

        # db-stored (user-created) object types
        obj_db = await storage.get_object_type(obj_type)
        try:
            type_def = hlp.type_def_from_source(obj_db.source, obj_db.id, Generic)
            object_actions_dict[obj_type] = otu.object_actions(TYPE_TO_PLUGIN, type_def, obj_db.source)
        except hlp.TypeDefException as e:
            await glob.logger.error(e)

    # add actions from ancestors
    for obj_type in valid_types.keys():
        otu.add_ancestor_actions(obj_type, object_actions_dict, glob.OBJECT_TYPES)

    # get services' actions
    for service_type, service_meta in valid_service_types().items():

        if service_meta.built_in:
            continue

        srv_type = await storage.get_service_type(service_type)
        try:
            srv_type_def = hlp.type_def_from_source(srv_type.source, service_type, Service)
            object_actions_dict[service_type] = otu.object_actions(TYPE_TO_PLUGIN, srv_type_def, srv_type.source)
        except Arcor2Exception:
            await glob.logger.exception(f"Error while processing service type {service_type}")

    glob.ACTIONS = object_actions_dict

    await notif.broadcast_event(events.ObjectTypesChangedEvent(data=list(object_actions_dict.keys())))


async def get_robot_instance(robot_id: str, end_effector_id: Optional[str] = None) -> Union[Robot, RobotService]:

    if robot_id in glob.SCENE_OBJECT_INSTANCES:
        robot_inst = glob.SCENE_OBJECT_INSTANCES[robot_id]
        if not isinstance(robot_inst, Robot):
            raise Arcor2Exception("Not a robot.")
        if end_effector_id and end_effector_id not in await hlp.run_in_executor(robot_inst.get_end_effectors_ids):
            raise Arcor2Exception("Unknown end effector ID.")
        return robot_inst
    else:
        robot_srv_inst = find_robot_service()
        if not robot_srv_inst or robot_id not in await hlp.run_in_executor(robot_srv_inst.get_robot_ids):
            raise Arcor2Exception("Unknown robot ID.")
        if end_effector_id and end_effector_id not in await hlp.run_in_executor(robot_srv_inst.get_end_effectors_ids,
                                                                                robot_id):
            raise Arcor2Exception("Unknown end effector ID.")
        return robot_srv_inst


async def execute_action(action_method: Callable, params: Dict[str, Any]) -> None:

    assert glob.RUNNING_ACTION

    await notif.broadcast_event(events.ActionExecutionEvent(data=events.ActionExecutionData(glob.RUNNING_ACTION)))

    evt = events.ActionResultEvent()
    evt.data.action_id = glob.RUNNING_ACTION

    try:
        action_result = await hlp.run_in_executor(action_method, *params.values())
    except Arcor2Exception as e:
        await glob.logger.error(e)
        evt.data.error = e.message
    except (AttributeError, TypeError) as e:
        await glob.logger.error(e)
        evt.data.error = str(e)
    else:
        if action_result is not None:
            try:
                evt.data.result = TYPE_TO_PLUGIN[type(action_result)].value_to_json(action_result)
            except KeyError:
                # temporal workaround for unsupported types
                evt.data.result = str(action_result)

    if glob.RUNNING_ACTION is None:
        # action was cancelled, do not send any event
        return

    await notif.broadcast_event(evt)
    glob.RUNNING_ACTION = None
    glob.RUNNING_ACTION_PARAMS = None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import argparse
import asyncio
import base64
import functools
import json
import os
import shutil
import sys
import tempfile
import zipfile
from typing import Optional, Set, Union, Awaitable, List
from datetime import datetime, timezone

from aiorun import run  # type: ignore
import websockets
from aiologger import Logger  # type: ignore
from aiologger.levels import LogLevel  # type: ignore
from dataclasses_jsonschema import ValidationError
from websockets.server import WebSocketServerProtocol as WsClient

import arcor2
from arcor2.exceptions import Arcor2Exception
from arcor2.data import rpc, compile_json_schemas
from arcor2.data.common import PackageState, PackageStateEnum, Project
from arcor2.data.execution import PackageMeta
from arcor2.data.events import ActionStateEvent, CurrentActionEvent, Event, PackageStateEvent, PackageInfoEvent
from arcor2.data.helpers import EVENT_MAPPING
from arcor2.helpers import RPC_DICT_TYPE, aiologger_formatter, server
from arcor2.settings import PROJECT_PATH
from arcor2.source.utils import make_executable


PORT = 6790

logger = Logger.with_default_handlers(name='manager', formatter=aiologger_formatter())

PROCESS: Union[asyncio.subprocess.Process, None] = None
PROJECT_EVENT: PackageStateEvent = PackageStateEvent()
PACKAGE_INFO_EVENT: Optional[PackageInfoEvent] = None
ACTION_EVENT: Optional[ActionStateEvent] = None
ACTION_ARGS_EVENT: Optional[CurrentActionEvent] = None
TASK = None

CLIENTS: Set = set()

MAIN_SCRIPT_NAME = "script.py"


def process_running() -> bool:

    return PROCESS is not None and PROCESS.returncode is None


async def project_state(event: PackageStateEvent):

    global PROJECT_EVENT
    PROJECT_EVENT = event
    await send_to_clients(event)


async def read_proc_stdout() -> None:

    global PROJECT_EVENT
    global ACTION_EVENT
    global ACTION_ARGS_EVENT
    global PACKAGE_INFO_EVENT

    logger.info("Reading script stdout...")

    assert PROCESS is not None
    assert PROCESS.stdout is not None

    await project_state(PackageStateEvent(data=PackageState(PackageStateEnum.RUNNING)))

    while process_running():
        try:
            stdout = await PROCESS.stdout.readuntil()
        except asyncio.exceptions.IncompleteReadError:
            break

        stripped = stdout.decode("utf-8").strip()

        try:
            data = json.loads(stripped)
        except json.decoder.JSONDecodeError as e:
            await logger.error(f"Script printed out: {stripped}")
            await logger.debug(e)
            continue

        if not isinstance(data, dict) or "event" not in data:
            await logger.error("Strange data from script: {}".format(data))
            continue

        try:
            evt = EVENT_MAPPING[data["event"]].from_dict(data)
        except ValidationError as e:
            await logger.error("Invalid event: {}, error: {}".format(data, e))
            continue

        if isinstance(evt, PackageStateEvent):
            await project_state(evt)
            continue
        elif isinstance(evt, ActionStateEvent):
            ACTION_EVENT = evt
        elif isinstance(evt, CurrentActionEvent):
            ACTION_ARGS_EVENT = evt
        elif isinstance(evt, PackageInfoEvent):
            PACKAGE_INFO_EVENT = evt

        await send_to_clients(evt)

    ACTION_EVENT = None
    ACTION_ARGS_EVENT = None
    PACKAGE_INFO_EVENT = None

    await project_state(PackageStateEvent(data=PackageState(PackageStateEnum.STOPPED)))
    logger.info(f"Process finished with returncode {PROCESS.returncode}.")


async def run_package_cb(req: rpc.execution.RunPackageRequest, ui: WsClient) -> None:

    global PROCESS
    global TASK

    if process_running():
        raise Arcor2Exception("Already running!")

    package_path = os.path.join(PROJECT_PATH, req.args.id)

    try:
        os.chdir(package_path)
    except FileNotFoundError:
        raise Arcor2Exception("Not found.")

    script_path = os.path.join(package_path, MAIN_SCRIPT_NAME)

    try:
        make_executable(script_path)
    except FileNotFoundError:
        raise Arcor2Exception("Not an execution package.")

    await logger.info(f"Starting script: {script_path}")
    PROCESS = await asyncio.create_subprocess_exec(script_path, stdin=asyncio.subprocess.PIPE,
                                                   stdout=asyncio.subprocess.PIPE,
                                                   stderr=asyncio.subprocess.STDOUT)
    if PROCESS.returncode is not None:
        raise Arcor2Exception("Failed to start project.")

    TASK = asyncio.ensure_future(read_proc_stdout())  # run task in background


async def stop_package_cb(req: rpc.execution.StopPackageRequest, ui: WsClient) -> None:

    global PACKAGE_INFO_EVENT

    if not process_running():
        raise Arcor2Exception("Project not running.")

    assert PROCESS is not None
    assert TASK is not None

    await logger.info("Terminating process")
    PROCESS.terminate()
    await logger.info("Waiting for process to finish...")
    await asyncio.wait([TASK])
    PACKAGE_INFO_EVENT = None


async def pause_package_cb(req: rpc.execution.PausePackageRequest, ui: WsClient) -> None:

    if not process_running():
        raise Arcor2Exception("Project not running.")

    assert PROCESS is not None
    assert PROCESS.stdin is not None

    if PROJECT_EVENT.data.state != PackageStateEnum.RUNNING:
        raise Arcor2Exception("Cannot pause.")

    PROCESS.stdin.write("p\n".encode())
    await PROCESS.stdin.drain()
    return None


async def resume_package_cb(req: rpc.execution.ResumePackageRequest, ui: WsClient) -> None:

    if not process_running():
        raise Arcor2Exception("Project not running.")

    assert PROCESS is not None and PROCESS.stdin is not None

    if PROJECT_EVENT.data.state != PackageStateEnum.PAUSED:
        raise Arcor2Exception("Cannot resume.")

    PROCESS.stdin.write("r\n".encode())
    await PROCESS.stdin.drain()
    return None


async def package_state_cb(req: rpc.execution.PackageStateRequest, ui: WsClient) ->\
        rpc.execution.PackageStateResponse:

    resp = rpc.execution.PackageStateResponse()
    resp.data.project = PROJECT_EVENT.data
    if ACTION_EVENT:
        resp.data.action = ACTION_EVENT.data
    if ACTION_ARGS_EVENT:
        resp.data.action_args = ACTION_ARGS_EVENT.data
    return resp


async def _upload_package_cb(req: rpc.execution.UploadPackageRequest, ui: WsClient) -> None:

    target_path = os.path.join(PROJECT_PATH, req.args.id)

    # TODO do not allow if there are manual changes?

    with tempfile.TemporaryDirectory() as tmpdirname:

        zip_path = os.path.join(tmpdirname, "publish.zip")

        b64_bytes = req.args.data.encode()
        zip_content = base64.b64decode(b64_bytes)

        with open(zip_path, "wb") as zip_file:
            zip_file.write(zip_content)

        try:
            with zipfile.ZipFile(zip_path, 'r') as zip_ref:
                zip_ref.extractall(tmpdirname)
        except zipfile.BadZipFile as e:
            await logger.error(e)
            raise Arcor2Exception("Invalid zip file.")

        os.remove(zip_path)

        try:
            shutil.rmtree(target_path)
        except FileNotFoundError:
            pass
        shutil.copytree(tmpdirname, target_path)

    script_path = os.path.join(target_path, MAIN_SCRIPT_NAME)
    make_executable(script_path)
    return None


async def read_package_meta(package_id: str) -> PackageMeta:

    target_path = os.path.join(PROJECT_PATH, package_id, "package.json")

    try:
        with open(target_path) as pkg_file:
            return PackageMeta.from_json(pkg_file.read())
    except (IOError, ValidationError) as e:
        asyncio.ensure_future(logger.error(f"Failed to read package.json of {package_id}: {e}"))
        return PackageMeta("N/A", datetime.fromtimestamp(0, tz=timezone.utc))


async def list_packages_cb(req: rpc.execution.ListPackagesRequest, ui: WsClient) ->\
        rpc.execution.ListPackagesResponse:

    resp = rpc.execution.ListPackagesResponse()

    subfolders = [f.path for f in os.scandir(PROJECT_PATH) if f.is_dir()]

    for folder_path in subfolders:

        if not os.path.isfile(os.path.join(folder_path, MAIN_SCRIPT_NAME)):
            continue

        package_dir = os.path.basename(folder_path)

        with open(os.path.join(folder_path, "data", "project.json")) as project_file:
            try:
                project = Project.from_json(project_file.read())
            except ValidationError as e:
                await logger.error(f"Failed to parse project file of {package_dir}: {e}")
                continue

        assert project.modified

        resp.data.append(rpc.execution.PackageSummary(package_dir, project.id, project.modified,
                                                      await read_package_meta(package_dir)))

        # TODO report manual changes (check last modification of files)?

    return resp


async def delete_package_cb(req: rpc.execution.DeletePackageRequest, ui: WsClient) -> None:

    if PACKAGE_INFO_EVENT and PACKAGE_INFO_EVENT.data and PACKAGE_INFO_EVENT.data.package_id == req.args.id:
        raise Arcor2Exception("Package is being executed.")

    target_path = os.path.join(PROJECT_PATH, req.args.id)

    try:
        shutil.rmtree(target_path)
    except FileNotFoundError:
        raise Arcor2Exception("Not found.")

    return None


async def rename_package_cb(req: rpc.execution.RenamePackageRequest, ui: WsClient) -> None:

    target_path = os.path.join(PROJECT_PATH, req.args.package_id, "package.json")

    pm = await read_package_meta(req.args.package_id)
    pm.name = req.args.new_name

    with open(target_path, "w") as pkg_file:
        pkg_file.write(pm.to_json())


async def send_to_clients(event: Event) -> None:

    if CLIENTS:
        data = event.to_json()
        await asyncio.wait([client.send(data) for client in CLIENTS])


async def register(websocket: WsClient) -> None:

    await logger.info("Registering new client")
    CLIENTS.add(websocket)

    tasks: List[Awaitable] = [websocket.send(PROJECT_EVENT.to_json())]

    if PACKAGE_INFO_EVENT:
        tasks.append(websocket.send(PACKAGE_INFO_EVENT.to_json()))

    await asyncio.gather(*tasks)


async def unregister(websocket: WsClient) -> None:
    await logger.info("Unregistering client")
    CLIENTS.remove(websocket)

RPC_DICT: RPC_DICT_TYPE = {
    rpc.execution.RunPackageRequest: run_package_cb,
    rpc.execution.StopPackageRequest: stop_package_cb,
    rpc.execution.PausePackageRequest: pause_package_cb,
    rpc.execution.ResumePackageRequest: resume_package_cb,
    rpc.execution.PackageStateRequest: package_state_cb,
    rpc.execution.UploadPackageRequest: _upload_package_cb,
    rpc.execution.ListPackagesRequest: list_packages_cb,
    rpc.execution.DeletePackageRequest: delete_package_cb,
    rpc.execution.RenamePackageRequest: rename_package_cb
}


async def aio_main() -> None:

    await websockets.serve(
        functools.partial(server, logger=logger, register=register, unregister=unregister, rpc_dict=RPC_DICT),
        '0.0.0.0', PORT)


def main() -> None:

    assert sys.version_info >= (3, 8)

    parser = argparse.ArgumentParser()

    parser.add_argument("-v", "--verbose", help="Increase output verbosity",
                        action="store_const", const=LogLevel.DEBUG, default=LogLevel.INFO)
    parser.add_argument('--version', action='version', version=arcor2.version(),
                        help="Shows ARCOR2 version and exits.")
    parser.add_argument('--api_version', action='version', version=arcor2.api_version(),
                        help="Shows API version and exits.")
    parser.add_argument("-a", "--asyncio_debug", help="Turn on asyncio debug mode.",
                        action="store_const", const=True, default=False)

    args = parser.parse_args()
    logger.level = args.verbose

    loop = asyncio.get_event_loop()
    loop.set_debug(enabled=args.asyncio_debug)

    compile_json_schemas()

    run(aio_main(), loop=loop, stop_on_unhandled_errors=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
from arcor2.object_types.upload import upload, UploadException


def main() -> None:

    try:
        arg = sys.argv[1]
    except IndexError:
        print("Usage: arcor2_upload_object_type.py package.module/Cls")
        return

    try:
        upload(arg)
    except UploadException as e:
        print(e)
        return
    print('Done')


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from arcor2.services.upload import upload_cls
import arcor2.service_types_utils as stu


def main():

    for _, srv_def in stu.built_in_services():
        upload_cls(srv_def)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import Set, Optional
import asyncio
import functools
from contextlib import asynccontextmanager
import copy

import quaternion  # type: ignore
from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import aio_persistent_storage as storage, helpers as hlp
from arcor2.data import rpc, events
from arcor2.data import common, object_type
from arcor2.exceptions import Arcor2Exception

from arcor2.server.robot import get_end_effector_pose
from arcor2.server.decorators import scene_needed, no_project, no_scene
from arcor2.server import globals as glob, notifications as notif
from arcor2.server.robot import collision
from arcor2.server.scene import add_object_to_scene, auto_add_object_to_scene, open_scene, add_service_to_scene,\
    clear_scene, get_instance
from arcor2.server.project import scene_object_pose_updated, remove_object_references_from_projects,\
    projects_using_object, associated_projects


OBJECTS_WITH_UPDATED_POSE: Set[str] = set()


@asynccontextmanager
async def managed_scene(scene_id: str, make_copy: bool = False):

    save_back = False

    if glob.SCENE and glob.SCENE.id == scene_id:
        if make_copy:
            scene = copy.deepcopy(glob.SCENE)
            save_back = True
        else:
            scene = glob.SCENE
    else:
        save_back = True
        scene = await storage.get_scene(scene_id)

    if make_copy:
        scene.id = common.uid()

    try:
        yield scene
    finally:
        if save_back:
            asyncio.ensure_future(storage.update_scene(scene))


@no_scene
async def new_scene_cb(req: rpc.scene.NewSceneRequest, ui: WsClient) -> None:
    """
    Creates and opens a new scene on the server. Fails if any scene is open or if scene id/name already exists.
    :param req:
    :return:
    """

    if glob.PACKAGE_STATE.state != common.PackageStateEnum.STOPPED:
        raise Arcor2Exception("Can't create scene while package runs.")

    assert glob.SCENE is None

    for scene_id in (await storage.get_scenes()).items:
        if req.args.name == scene_id.name:
            raise Arcor2Exception("Name already used.")

    if req.dry_run:
        return None

    glob.SCENE = common.Scene(common.uid(), req.args.name, desc=req.args.desc)
    asyncio.ensure_future(notif.broadcast_event(events.OpenScene(data=events.OpenSceneData(glob.SCENE))))
    return None


@scene_needed
@no_project
async def close_scene_cb(req: rpc.scene.CloseSceneRequest, ui: WsClient) -> None:
    """
    Closes scene on the server.
    :param req:
    :return:
    """

    assert glob.SCENE

    if not req.args.force and glob.SCENE.has_changes():
        raise Arcor2Exception("Scene has unsaved changes.")

    if req.dry_run:
        return None

    await clear_scene()
    OBJECTS_WITH_UPDATED_POSE.clear()
    asyncio.ensure_future(notif.broadcast_event(events.SceneClosed()))
    return None


@scene_needed
@no_project
async def save_scene_cb(req: rpc.scene.SaveSceneRequest, ui: WsClient) -> None:

    assert glob.SCENE
    await storage.update_scene(glob.SCENE)
    glob.SCENE.modified = (await storage.get_scene(glob.SCENE.id)).modified
    asyncio.ensure_future(notif.broadcast_event(events.SceneSaved()))
    for obj_id in OBJECTS_WITH_UPDATED_POSE:
        asyncio.ensure_future(scene_object_pose_updated(glob.SCENE.id, obj_id))
    OBJECTS_WITH_UPDATED_POSE.clear()
    return None


@no_project
async def open_scene_cb(req: rpc.scene.OpenSceneRequest, ui: WsClient) -> None:

    if glob.PACKAGE_STATE.state != common.PackageStateEnum.STOPPED:
        raise Arcor2Exception("Can't open scene while package runs.")

    await open_scene(req.args.id)
    assert glob.SCENE
    asyncio.ensure_future(notif.broadcast_event(events.OpenScene(data=events.OpenSceneData(glob.SCENE))))
    return None


async def list_scenes_cb(req: rpc.scene.ListScenesRequest, ui: WsClient) -> rpc.scene.ListScenesResponse:

    scenes = await storage.get_scenes()
    return rpc.scene.ListScenesResponse(data=scenes.items)


@scene_needed
@no_project
async def add_object_to_scene_cb(req: rpc.scene.AddObjectToSceneRequest, ui: WsClient) -> None:

    assert glob.SCENE

    obj = common.SceneObject(common.uid(), req.args.name, req.args.type, req.args.pose)

    await add_object_to_scene(obj, dry_run=req.dry_run)

    if req.dry_run:
        return None

    glob.SCENE.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.SceneObjectChanged(events.EventType.ADD, data=obj)))
    return None


@scene_needed
@no_project
async def auto_add_object_to_scene_cb(req: rpc.scene.AutoAddObjectToSceneRequest, ui: WsClient) -> None:
    assert glob.SCENE

    obj = req.args
    await auto_add_object_to_scene(obj.type, req.dry_run)

    if req.dry_run:
        return None

    glob.SCENE.update_modified()
    return None


@scene_needed
@no_project
async def add_service_to_scene_cb(req: rpc.scene.AddServiceToSceneRequest, ui: WsClient) -> None:

    assert glob.SCENE

    srv = req.args
    await add_service_to_scene(srv, req.dry_run)

    if req.dry_run:
        return None

    glob.SCENE.services.append(srv)
    glob.SCENE.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.SceneServiceChanged(events.EventType.ADD, data=srv)))
    return None


@scene_needed
async def scene_object_usage_request_cb(req: rpc.scene.SceneObjectUsageRequest, ui: WsClient) -> \
        rpc.scene.SceneObjectUsageResponse:
    """
    Works for both services and objects.
    :param req:
    :return:
    """

    assert glob.SCENE

    if not (any(obj.id == req.args.id for obj in glob.SCENE.objects) or
            any(srv.type == req.args.id for srv in glob.SCENE.services)):
        raise Arcor2Exception("Unknown ID.")

    resp = rpc.scene.SceneObjectUsageResponse()

    async for project in projects_using_object(glob.SCENE.id, req.args.id):
        resp.data.add(project.id)

    return resp


# TODO move to objects
@scene_needed
async def action_param_values_cb(req: rpc.objects.ActionParamValuesRequest, ui: WsClient) -> \
        rpc.objects.ActionParamValuesResponse:

    inst = get_instance(req.args.id)

    parent_params = {}

    for pp in req.args.parent_params:
        parent_params[pp.id] = pp.value

    try:
        method_name, required_parent_params = inst.DYNAMIC_PARAMS[req.args.param_id]
    except KeyError:
        raise Arcor2Exception("Unknown parameter or values not constrained.")

    if parent_params.keys() != required_parent_params:
        raise Arcor2Exception("Not all required parent params were given.")

    # TODO validate method parameters vs parent_params (check types)?

    resp = rpc.objects.ActionParamValuesResponse()

    try:
        method = getattr(inst, method_name)
    except AttributeError:
        await glob.logger.error(f"Unable to get values for parameter {req.args.param_id}, "
                                f"object/service {inst.id} has no method named {method_name}.")
        raise Arcor2Exception("System error.")

    # TODO update hlp.run_in_executor to support kwargs
    resp.data = await asyncio.get_event_loop().run_in_executor(None, functools.partial(method, **parent_params))
    return resp


@scene_needed
@no_project
async def remove_from_scene_cb(req: rpc.scene.RemoveFromSceneRequest, ui: WsClient) -> None:

    assert glob.SCENE

    if not req.args.force and {proj.name async for proj in projects_using_object(glob.SCENE.id, req.args.id)}:
        raise Arcor2Exception("Can't remove object/service that is used in project(s).")

    if req.dry_run:
        return None

    if req.args.id in glob.SCENE_OBJECT_INSTANCES:

        obj = glob.SCENE.object(req.args.id)
        glob.SCENE.objects = [obj for obj in glob.SCENE.objects if obj.id != req.args.id]
        obj_inst = glob.SCENE_OBJECT_INSTANCES[req.args.id]
        await collision(obj_inst, remove=True)
        del glob.SCENE_OBJECT_INSTANCES[req.args.id]
        if req.args.id in OBJECTS_WITH_UPDATED_POSE:
            OBJECTS_WITH_UPDATED_POSE.remove(req.args.id)
        asyncio.ensure_future(notif.broadcast_event(events.SceneObjectChanged(events.EventType.REMOVE, data=obj)))

    elif req.args.id in glob.SERVICES_INSTANCES:

        # first check if some object is not using it
        for obj in glob.SCENE.objects:
            if req.args.id in glob.OBJECT_TYPES[obj.type].needs_services:
                raise Arcor2Exception(f"Object {obj.id} ({obj.type}) "
                                      f"relies on the service to be removed: {req.args.id}.")

        srv = glob.SCENE.service(req.args.id)
        glob.SCENE.services = [srv for srv in glob.SCENE.services if srv.type != req.args.id]
        del glob.SERVICES_INSTANCES[req.args.id]
        asyncio.ensure_future(notif.broadcast_event(events.SceneServiceChanged(events.EventType.REMOVE, data=srv)))

    else:
        raise Arcor2Exception("Unknown id.")

    glob.SCENE.update_modified()
    asyncio.ensure_future(remove_object_references_from_projects(req.args.id))
    return None


@scene_needed
@no_project
async def update_object_pose_using_robot_cb(req: rpc.objects.UpdateObjectPoseUsingRobotRequest, ui: WsClient) -> \
        None:
    """
    Updates object's pose using a pose of the robot's end effector.
    :param req:
    :return:
    """

    assert glob.SCENE

    if req.args.id == req.args.robot.robot_id:
        raise Arcor2Exception("Robot cannot update its own pose.")

    scene_object = glob.SCENE.object(req.args.id)

    if glob.OBJECT_TYPES[scene_object.type].needs_services:
        raise Arcor2Exception("Can't manipulate object created by service.")

    obj_inst = glob.SCENE_OBJECT_INSTANCES[req.args.id]

    if obj_inst.collision_model:
        if isinstance(obj_inst.collision_model, object_type.Mesh) and req.args.pivot != rpc.objects.PivotEnum.MIDDLE:
            raise Arcor2Exception("Only middle pivot point is supported for objects with mesh collision model.")
    elif req.args.pivot != rpc.objects.PivotEnum.MIDDLE:
        raise Arcor2Exception("Only middle pivot point is supported for objects without collision model.")

    new_pose = await get_end_effector_pose(req.args.robot.robot_id, req.args.robot.end_effector)

    position_delta = common.Position()

    if obj_inst.collision_model:
        if isinstance(obj_inst.collision_model, object_type.Box):
            if req.args.pivot == rpc.objects.PivotEnum.TOP:
                position_delta.z -= obj_inst.collision_model.size_z / 2
            elif req.args.pivot == rpc.objects.PivotEnum.BOTTOM:
                position_delta.z += obj_inst.collision_model.size_z / 2
        elif isinstance(obj_inst.collision_model, object_type.Cylinder):
            if req.args.pivot == rpc.objects.PivotEnum.TOP:
                position_delta.z -= obj_inst.collision_model.height / 2
            elif req.args.pivot == rpc.objects.PivotEnum.BOTTOM:
                position_delta.z += obj_inst.collision_model.height / 2
        elif isinstance(obj_inst.collision_model, object_type.Sphere):
            if req.args.pivot == rpc.objects.PivotEnum.TOP:
                position_delta.z -= obj_inst.collision_model.radius / 2
            elif req.args.pivot == rpc.objects.PivotEnum.BOTTOM:
                position_delta.z += obj_inst.collision_model.radius / 2

    position_delta = position_delta.rotated(new_pose.orientation)

    scene_object.pose.position.x = new_pose.position.x - position_delta.x
    scene_object.pose.position.y = new_pose.position.y - position_delta.y
    scene_object.pose.position.z = new_pose.position.z - position_delta.z

    scene_object.pose.orientation.set_from_quaternion(
        new_pose.orientation.as_quaternion()*quaternion.quaternion(0, 1, 0, 0))
    obj_inst.pose = scene_object.pose

    glob.SCENE.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.SceneObjectChanged(events.EventType.UPDATE, data=scene_object)))
    OBJECTS_WITH_UPDATED_POSE.add(scene_object.id)
    return None


@scene_needed
@no_project
async def update_object_pose_cb(req: rpc.scene.UpdateObjectPoseRequest, ui: WsClient) -> None:

    assert glob.SCENE

    obj = glob.SCENE.object(req.args.object_id)

    if glob.OBJECT_TYPES[obj.type].needs_services:
        raise Arcor2Exception("Can't manipulate object created by service.")

    obj.pose = req.args.pose
    glob.SCENE_OBJECT_INSTANCES[req.args.object_id].pose = req.args.pose

    glob.SCENE.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.SceneObjectChanged(events.EventType.UPDATE, data=obj)))

    OBJECTS_WITH_UPDATED_POSE.add(obj.id)
    return None


@scene_needed
@no_project
async def rename_object_cb(req: rpc.scene.RenameObjectRequest, ui: WsClient) -> None:

    assert glob.SCENE

    target_obj = glob.SCENE.object(req.args.id)

    for obj_name in glob.SCENE.object_names():
        if obj_name == req.args.new_name:
            raise Arcor2Exception("Object name already exists.")

    if not hlp.is_valid_identifier(req.args.new_name):
        raise Arcor2Exception("Object name invalid (should be snake_case).")

    if req.dry_run:
        return None

    target_obj.name = req.args.new_name

    glob.SCENE.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.SceneObjectChanged(events.EventType.UPDATE, data=target_obj)))
    return None


async def rename_scene_cb(req: rpc.scene.RenameSceneRequest, ui: WsClient) -> None:

    # TODO unique_name(req.args.new_name, (await scene_names()))

    async with managed_scene(req.args.id) as scene:
        scene.name = req.args.new_name
        asyncio.ensure_future(notif.broadcast_event(events.SceneChanged(events.EventType.UPDATE_BASE,
                                                                        data=scene.bare())))
    return None


@no_scene
async def delete_scene_cb(req: rpc.scene.DeleteSceneRequest, ui: WsClient) -> \
        Optional[rpc.scene.DeleteSceneResponse]:

    assoc_projects = await associated_projects(req.args.id)

    if assoc_projects:
        resp = rpc.scene.DeleteSceneResponse(result=False)
        resp.messages = ["Scene has associated projects."]
        resp.data = assoc_projects
        return resp

    if req.dry_run:
        return None

    scene = await storage.get_scene(req.args.id)
    await storage.delete_scene(req.args.id)
    asyncio.ensure_future(notif.broadcast_event(events.SceneChanged(events.EventType.REMOVE,
                                                                    data=scene.bare())))
    return None


async def projects_with_scene_cb(req: rpc.scene.ProjectsWithSceneRequest, ui: WsClient) -> \
        rpc.scene.ProjectsWithSceneResponse:

    resp = rpc.scene.ProjectsWithSceneResponse()
    resp.data = await associated_projects(req.args.id)
    return resp


async def update_scene_description_cb(req: rpc.scene.UpdateSceneDescriptionRequest, ui: WsClient) -> None:

    async with managed_scene(req.args.scene_id) as scene:
        scene.desc = req.args.new_description
        scene.update_modified()
        asyncio.ensure_future(notif.broadcast_event(events.SceneChanged(events.EventType.UPDATE_BASE,
                                                                        data=scene.bare())))
    return None


@scene_needed
async def update_service_configuration_cb(req: rpc.scene.UpdateServiceConfigurationRequest, ui: WsClient) -> None:

    assert glob.SCENE

    srv = glob.SCENE.service(req.args.type)

    # first check if some object is not using it
    for obj in glob.SCENE.objects:
        if req.args.type in glob.OBJECT_TYPES[obj.type].needs_services:
            raise Arcor2Exception(f"Object {obj.name} ({obj.type}) relies "
                                  f"on the service to be removed: {req.args.type}.")

    if req.dry_run:
        return None

    # TODO destroy current instance
    glob.SERVICES_INSTANCES[req.args.type] = await hlp.run_in_executor(glob.TYPE_DEF_DICT[req.args.type],
                                                                       req.args.new_configuration)
    srv.configuration_id = req.args.new_configuration

    glob.SCENE.update_modified()
    asyncio.ensure_future(notif.broadcast_event(events.SceneServiceChanged(events.EventType.UPDATE, data=srv)))

    return None


async def copy_scene_cb(req: rpc.scene.CopySceneRequest, ui: WsClient) -> None:

    # TODO check if target_name is unique
    async with managed_scene(req.args.source_id, make_copy=True) as scene:
        scene.name = req.args.target_name
        asyncio.ensure_future(notif.broadcast_event(events.SceneChanged(events.EventType.UPDATE_BASE,
                                                                        data=scene.bare())))

    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import asyncio
from typing import List, Dict

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2.source.object_types import new_object_type_source
from arcor2 import object_types_utils as otu, helpers as hlp
from arcor2.data.common import Position, Pose
from arcor2.data.object_type import Model3dType, MeshFocusAction
from arcor2.data import rpc, events
from arcor2 import aio_persistent_storage as storage
from arcor2.object_types import Generic
from arcor2.exceptions import Arcor2Exception
from arcor2.parameter_plugins import TYPE_TO_PLUGIN

from arcor2.server.decorators import scene_needed, no_project
from arcor2.server import objects_services_actions as osa, notifications as notif, globals as glob
from arcor2.server.robot import get_end_effector_pose
from arcor2.server.project import scene_object_pose_updated


FOCUS_OBJECT: Dict[str, Dict[int, Pose]] = {}  # object_id / idx, pose
FOCUS_OBJECT_ROBOT: Dict[str, rpc.common.RobotArg] = {}  # key: object_id


def clean_up_after_focus(obj_id: str) -> None:

    try:
        del FOCUS_OBJECT[obj_id]
    except KeyError:
        pass

    try:
        del FOCUS_OBJECT_ROBOT[obj_id]
    except KeyError:
        pass


@scene_needed
@no_project
async def focus_object_start_cb(req: rpc.objects.FocusObjectStartRequest, ui: WsClient) -> None:

    obj_id = req.args.object_id

    if obj_id in FOCUS_OBJECT_ROBOT:
        raise Arcor2Exception("Focusing already started.")

    if obj_id not in glob.SCENE_OBJECT_INSTANCES:
        raise Arcor2Exception("Unknown object.")

    inst = await osa.get_robot_instance(req.args.robot.robot_id, req.args.robot.end_effector)

    if not glob.ROBOT_META[inst.__class__.__name__].features.focus:
        raise Arcor2Exception("Robot/service does not support focusing.")

    obj_type = glob.OBJECT_TYPES[osa.get_obj_type_name(obj_id)]

    if not obj_type.object_model or obj_type.object_model.type != Model3dType.MESH:
        raise Arcor2Exception("Only available for objects with mesh model.")

    assert obj_type.object_model.mesh

    focus_points = obj_type.object_model.mesh.focus_points

    if not focus_points:
        raise Arcor2Exception("focusPoints not defined for the mesh.")

    FOCUS_OBJECT_ROBOT[req.args.object_id] = req.args.robot
    FOCUS_OBJECT[obj_id] = {}
    await glob.logger.info(f'Start of focusing for {obj_id}.')
    return None


@no_project
async def focus_object_cb(req: rpc.objects.FocusObjectRequest, ui: WsClient) -> rpc.objects.FocusObjectResponse:

    obj_id = req.args.object_id
    pt_idx = req.args.point_idx

    if obj_id not in glob.SCENE_OBJECT_INSTANCES:
        raise Arcor2Exception("Unknown object_id.")

    obj_type = glob.OBJECT_TYPES[osa.get_obj_type_name(obj_id)]

    assert obj_type.object_model and obj_type.object_model.mesh

    focus_points = obj_type.object_model.mesh.focus_points

    assert focus_points

    if pt_idx < 0 or pt_idx > len(focus_points)-1:
        raise Arcor2Exception("Index out of range.")

    if obj_id not in FOCUS_OBJECT:
        await glob.logger.info(f'Start of focusing for {obj_id}.')
        FOCUS_OBJECT[obj_id] = {}

    robot_id, end_effector = FOCUS_OBJECT_ROBOT[obj_id].as_tuple()

    FOCUS_OBJECT[obj_id][pt_idx] = await get_end_effector_pose(robot_id, end_effector)

    r = rpc.objects.FocusObjectResponse()
    r.data.finished_indexes = list(FOCUS_OBJECT[obj_id].keys())
    return r


@scene_needed
@no_project
async def focus_object_done_cb(req: rpc.objects.FocusObjectDoneRequest, ui: WsClient) -> None:

    obj_id = req.args.id

    if obj_id not in FOCUS_OBJECT:
        raise Arcor2Exception("focusObjectStart/focusObject has to be called first.")

    obj_type = glob.OBJECT_TYPES[osa.get_obj_type_name(obj_id)]

    assert obj_type.object_model and obj_type.object_model.mesh

    focus_points = obj_type.object_model.mesh.focus_points

    assert focus_points

    if len(FOCUS_OBJECT[obj_id]) < len(focus_points):
        raise Arcor2Exception("Not all points were done.")

    robot_id, end_effector = FOCUS_OBJECT_ROBOT[obj_id].as_tuple()
    robot_inst = await osa.get_robot_instance(robot_id)
    assert hasattr(robot_inst, "focus")  # mypy does not deal with hasattr

    assert glob.SCENE

    obj = glob.SCENE.object(obj_id)

    fp: List[Position] = []
    rp: List[Position] = []

    for idx, pose in FOCUS_OBJECT[obj_id].items():

        fp.append(focus_points[idx].position)
        rp.append(pose.position)

    mfa = MeshFocusAction(fp, rp)

    await glob.logger.debug(f'Attempt to focus for object {obj_id}, data: {mfa}')

    try:
        obj.pose = await hlp.run_in_executor(robot_inst.focus, mfa)  # type: ignore
    except Arcor2Exception as e:
        await glob.logger.error(f"Focus failed with: {e}, mfa: {mfa}.")
        raise Arcor2Exception("Focusing failed.") from e

    await glob.logger.info(f"Done focusing for {obj_id}.")

    clean_up_after_focus(obj_id)

    asyncio.ensure_future(notif.broadcast_event(events.SceneObjectChanged(events.EventType.UPDATE, data=obj)))
    asyncio.ensure_future(scene_object_pose_updated(glob.SCENE.id, obj.id))
    return None


async def new_object_type_cb(req: rpc.objects.NewObjectTypeRequest, ui: WsClient) -> None:

    meta = req.args

    if meta.type in glob.OBJECT_TYPES:
        raise Arcor2Exception("Object type already exists.")

    if meta.base not in glob.OBJECT_TYPES:
        raise Arcor2Exception(f"Unknown base object type '{meta.base}', "
                              f"known types are: {', '.join(glob.OBJECT_TYPES.keys())}.")

    if not hlp.is_valid_type(meta.type):
        raise Arcor2Exception("Object type invalid (should be CamelCase).")

    if req.dry_run:
        return None

    obj = meta.to_object_type()
    obj.source = new_object_type_source(glob.OBJECT_TYPES[meta.base], meta)

    if meta.object_model and meta.object_model.type != Model3dType.MESH:
        assert meta.type == meta.object_model.model().id
        await storage.put_model(meta.object_model.model())

    # TODO check whether mesh id exists - if so, then use existing mesh, if not, upload a new one
    if meta.object_model and meta.object_model.type == Model3dType.MESH:
        # ...get whole mesh (focus_points) based on mesh id
        assert meta.object_model.mesh
        try:
            meta.object_model.mesh = await storage.get_mesh(meta.object_model.mesh.id)
        except storage.PersistentStorageException as e:
            await glob.logger.error(e)
            raise Arcor2Exception(f"Mesh ID {meta.object_model.mesh.id} does not exist.")

    await storage.update_object_type(obj)

    glob.OBJECT_TYPES[meta.type] = meta
    glob.ACTIONS[meta.type] = otu.object_actions(TYPE_TO_PLUGIN,
                                                 hlp.type_def_from_source(obj.source, obj.id, Generic), obj.source)
    otu.add_ancestor_actions(meta.type, glob.ACTIONS, glob.OBJECT_TYPES)

    asyncio.ensure_future(notif.broadcast_event(events.ObjectTypesChangedEvent(data=[meta.type])))
    return None


async def get_object_actions_cb(req: rpc.objects.GetActionsRequest, ui: WsClient) -> rpc.objects.GetActionsResponse:

    try:
        return rpc.objects.GetActionsResponse(data=glob.ACTIONS[req.args.type])
    except KeyError:
        raise Arcor2Exception(f"Unknown object type: '{req.args.type}'.")


async def get_object_types_cb(req: rpc.objects.GetObjectTypesRequest, ui: WsClient) ->\
        rpc.objects.GetObjectTypesResponse:
    return rpc.objects.GetObjectTypesResponse(data=list(glob.OBJECT_TYPES.values()))



from typing import Dict, Type, Set
import inspect
import pkgutil
import importlib

# TODO just temporary solution - allow loading plugins based on env. var or from project service?
from arcor2 import parameter_plugins
from arcor2.parameter_plugins.base import ParameterPlugin

PARAM_PLUGINS: Dict[str, Type[ParameterPlugin]] = {}

PLUGINS: Set[Type[ParameterPlugin]] = set()

for _, module_name, _ in pkgutil.iter_modules(parameter_plugins.__path__):  # type: ignore

    module = importlib.import_module(f"arcor2.parameter_plugins.{module_name}")

    for name, obj in inspect.getmembers(module):
        if not inspect.isclass(obj) or inspect.isabstract(obj) or not issubclass(obj, ParameterPlugin):
            continue
        PLUGINS.add(obj)


TYPE_TO_PLUGIN: Dict[Type, Type[ParameterPlugin]] = {}

for plug in PLUGINS:
    if plug.type_name() in PARAM_PLUGINS:
        print(f"Plugin for type {plug.type_name()} already registered.")
        continue
    # TODO figure out why mypy complains
    # Only concrete class can be given where "Type[ParameterPlugin]" is expected
    PARAM_PLUGINS[plug.type_name()] = plug  # type: ignore
    TYPE_TO_PLUGIN[plug.type()] = plug  # type: ignore
